#!/usr/bin/env python3

################################################
#
#   Object to update
#       MetaWorkflowRun[json]
#        and WorkflowRun[json]
#
#   Michele Berselli
#   berselli.michele@gmail.com
#
################################################

################################################
#   Libraries
################################################
import sys, os

################################################
#   RunUpdate
################################################
class RunUpdate(object):
    """Class to handle MetaWorkflowRun[obj] and WorkflowRun[obj] updates.
    """

    def __init__(self, wflrun_obj):
        """Constructor method.
        Initialize object and attributes.

        :param wflrun_obj: MetaWorkflowRun[obj] representing a MetaWorkflowRun[json]
        :type wflrun_obj: object
        """
        # Basic attributes
        self.wflrun_obj = wflrun_obj
    #end def

    def reset_steps(self, steps_name):
        """Reset WorkflowRun[obj] corresponding to step in steps_name.

        :param steps_name: List of names for steps that need to be reset
        :type steps_name: list(str)
        :return: Updated workflow_runs and final_status information
        :rtype: dict
        """
        for name in steps_name:
            self.wflrun_obj.reset_step(name)
        #end for
        return {'final_status':  self.wflrun_obj.update_status(),
                'workflow_runs': self.wflrun_obj.runs_to_json()}
    #end def

    def reset_shards(self, shards_name):
        """Reset WorkflowRun[obj] with shard_name in shards_name.

        :param shards_name: List of shard_names for WorkflowRun[obj] that need to be reset
        :type shards_name: list(str)
        :return: Updated workflow_runs and final_status information
        :rtype: dict
        """
        for name in shards_name:
            self.wflrun_obj.reset_shard(name)
        #end for
        return {'final_status':  self.wflrun_obj.update_status(),
                'workflow_runs': self.wflrun_obj.runs_to_json()}
    #end def

    def import_steps(self, wflrun_obj, steps_name, import_input=True):
        """Update current MetaWorkflowRun[obj] information.
        Import and use information from specified wflrun_obj.
        Update WorkflowRun[obj] up to steps specified by steps_name

        :param wflrun_obj: MetaWorkflowRun[obj] to import information from
        :type wflrun_obj: object
        :param steps_name: List of names for steps to import
        :type steps_name: list(str)
        :return: MetaWorkflowRun[json]
        :rtype: dict
        """
        ## Import input
        if import_input:
            self.wflrun_obj.input = wflrun_obj.input
        #end if
        ## Import WorkflowRun objects
        for name in steps_name:
            queue = [] # queue of steps to import
                       #    name step and its dependencies
            # Get workflow-runs corresponding to name step
            for shard_name, run_obj in self.wflrun_obj.runs.items():
                if name == shard_name.split(':')[0]:
                    queue.append(run_obj)
                #end if
            #end for
            # Iterate queue, get dependencies and import workflow-runs
            while queue:
                run_obj = queue.pop(0)
                shard_name = run_obj.shard_name
                dependencies = run_obj.dependencies
                try:
                    self.wflrun_obj.runs[shard_name] = wflrun_obj.runs[shard_name]
                except KeyError as e:
                    # raise ValueError('JSON content error, missing information for workflow-run "{0}"\n'
                    #                     .format(e.args[0]))
                    continue
                #end try
                for dependency in dependencies:
                    queue.append(self.wflrun_obj.runs[dependency])
                #end for
            #end while
        #end for
        # Update final_status
        self.wflrun_obj.update_status()

        return self.wflrun_obj.to_json()
    #end def

#end class



import click

from magma_smaht.create_metawfr import (
    mwfr_illumina_alignment,
    mwfr_pacbio_alignment,
    mwfr_fastqc,
    mwfr_hic_alignment,
    mwfr_ont_alignment,
)
from magma_smaht.utils import get_auth_key


@click.command()
@click.help_option("--help", "-h")
@click.option(
    "-f", "--fileset-accession", required=True, type=str, help="Fileset accession"
)
@click.option(
    "-l",
    "--length-required",
    required=True,
    type=int,
    help="Required length (can be obtained from FastQC output)",
)
@click.option(
    "-e",
    "--auth-env",
    required=True,
    type=str,
    help="Name of environment in smaht-keys file",
)
def mwfr_illumina_alignment_cmd(fileset_accession, length_required, auth_env):
    """Creates a MetaWorflowRun item in the portal for Illumina alignment of submitted files within a fileset"""
    smaht_key = get_auth_key(auth_env)
    mwfr_illumina_alignment(fileset_accession, length_required, smaht_key)


@click.command()
@click.help_option("--help", "-h")
@click.option(
    "-f", "--fileset-accession", required=True, type=str, help="Fileset accession"
)
@click.option(
    "-e",
    "--auth-env",
    required=True,
    type=str,
    help="Name of environment in smaht-keys file",
)
def mwfr_pacbio_alignment_cmd(fileset_accession, auth_env):
    """Creates a MetaWorflowRun item in the portal for PacBio alignment of submitted files within a fileset"""
    smaht_key = get_auth_key(auth_env)
    mwfr_pacbio_alignment(fileset_accession, smaht_key)


@click.command()
@click.help_option("--help", "-h")
@click.option(
    "-f", "--fileset-accession", required=True, type=str, help="Fileset accession"
)
@click.option(
    "-e",
    "--auth-env",
    required=True,
    type=str,
    help="Name of environment in smaht-keys file",
)
def mwfr_hic_alignment_cmd(fileset_accession, auth_env):
    """Creates a MetaWorflowRun item in the portal for HIC alignment of submitted files within a fileset"""
    smaht_key = get_auth_key(auth_env)
    mwfr_hic_alignment(fileset_accession, smaht_key)


@click.command()
@click.help_option("--help", "-h")
@click.option(
    "-f", "--fileset-accession", required=True, type=str, help="Fileset accession"
)
@click.option(
    "-e",
    "--auth-env",
    required=True,
    type=str,
    help="Name of environment in smaht-keys file",
)
def mwfr_ont_alignment_cmd(fileset_accession, auth_env):
    """Creates a MetaWorflowRun item in the portal for ONT alignment of submitted files within a fileset"""
    smaht_key = get_auth_key(auth_env)
    mwfr_ont_alignment(fileset_accession, smaht_key)



@click.command()
@click.help_option("--help", "-h")
@click.option(
    "-f", "--fileset-accession", required=True, type=str, help="Fileset accession"
)
@click.option(
    "-e",
    "--auth-env",
    required=True,
    type=str,
    help="Name of environment in smaht-keys file",
)
def mwfr_fastqc_cmd(fileset_accession, auth_env):
    smaht_key = get_auth_key(auth_env)
    mwfr_fastqc(fileset_accession, smaht_key)



#!/usr/bin/env python3

################################################
#
#   create_metawfr
#
################################################
from __future__ import annotations

import datetime
import json
import uuid
from functools import lru_cache
from typing import Any, List

from dcicutils import ff_utils

from magma_ff.metawfl import MetaWorkflow
from magma_ff.metawflrun import MetaWorkflowRun
from magma_ff.utils import JsonObject, keep_last_item, make_embed_request


COHORT_ANALYSIS_TYPE = "CohortAnalysis"
FILE_FORMAT = "file_format"
SAMPLE_PROCESSING_TYPE = "SampleProcessing"
SAMPLE_TYPE = "Sample"
SOMATIC_ANALYSIS_TYPE = "SomaticAnalysis"
TYPE = "@type"
UUID = "uuid"


def create_meta_workflow_run(
    item_identifier: str,
    meta_workflow_identifier: str,
    auth_key: JsonObject,
    post: bool = True,
    patch: bool = True,
) -> JsonObject:
    """Create a MetaWorkflowRun for the given item and MetaWorkflow.

    POST MetaWorkflowRun and PATCH associated item as instructed.

    :param item_identfier: Identifier (e.g. UUID, @id) for item from
        which to create the MetaWorkflowRun
    :param meta_workflow_identifier: Identifier for the MetaWorkflow
        from which to create the MetaWorkflowRun
    :param auth_key: Authorization keys for C4 account
    :param post: Whether to POST the MetaWorkflowRun created
    :param patch: Whether to PATCH the item given by the
        item_identifier with the created MetaWorkflowRun
    :returns: MetaWorkflowRun created
    """
    item = ff_utils.get_metadata(item_identifier, key=auth_key, add_on="frame=object")
    if _is_item_of_type(SAMPLE_TYPE, item):
        return create_meta_workflow_run_from_sample(
            item_identifier, meta_workflow_identifier, auth_key, post=post, patch=patch
        )
    if _is_item_of_type(SAMPLE_PROCESSING_TYPE, item):
        return create_meta_workflow_run_from_sample_processing(
            item_identifier, meta_workflow_identifier, auth_key, post=post, patch=patch
        )
    if _is_item_of_type(COHORT_ANALYSIS_TYPE, item):
        return create_meta_workflow_run_from_cohort_analysis(
            item_identifier, meta_workflow_identifier, auth_key, post=post, patch=patch
        )
    if _is_item_of_type(SOMATIC_ANALYSIS_TYPE, item):
        return create_meta_workflow_run_from_somatic_analysis(
            item_identifier, meta_workflow_identifier, auth_key, post=post, patch=patch
        )
    raise MetaWorkflowRunCreationError(
        f"No methods available to create MetaWorkflowRun for item of type(s):"
        f" {_get_item_types(item)}"
    )


def _is_item_of_type(item_type: str, item: JsonObject) -> bool:
    item_types = _get_item_types(item)
    if item_type in item_types:
        return True
    return False


def _get_item_types(item: JsonObject) -> List[str]:
    return item.get(TYPE, [])


def create_meta_workflow_run_from_sample(
    sample_identifier: str,
    meta_workflow_identifier: str,
    auth_key: JsonObject,
    post: bool = True,
    patch: bool = True,
) -> JsonObject:
    return _create_meta_workflow_run_for_item_type(
        MetaWorkflowRunFromSample,
        sample_identifier,
        meta_workflow_identifier,
        auth_key,
        post=post,
        patch=patch,
    )


def create_meta_workflow_run_from_sample_processing(
    sample_processing_identifier: str,
    meta_workflow_identifier: str,
    auth_key: JsonObject,
    post: bool = True,
    patch: bool = True,
) -> JsonObject:
    return _create_meta_workflow_run_for_item_type(
        MetaWorkflowRunFromSampleProcessing,
        sample_processing_identifier,
        meta_workflow_identifier,
        auth_key,
        post=post,
        patch=patch,
    )


def create_meta_workflow_run_from_cohort_analysis(
    cohort_analysis_identifier: str,
    meta_workflow_identifier: str,
    auth_key: JsonObject,
    post: bool = True,
    patch: bool = True,
) -> JsonObject:
    return _create_meta_workflow_run_for_item_type(
        MetaWorkflowRunFromCohortAnalysis,
        cohort_analysis_identifier,
        meta_workflow_identifier,
        auth_key,
        post=post,
        patch=patch,
    )


def create_meta_workflow_run_from_somatic_analysis(
    somatic_analysis_identifier: str,
    meta_workflow_identifier: str,
    auth_key: JsonObject,
    post: bool = True,
    patch: bool = True,
) -> JsonObject:
    return _create_meta_workflow_run_for_item_type(
        MetaWorkflowRunFromSomaticAnalysis,
        somatic_analysis_identifier,
        meta_workflow_identifier,
        auth_key,
        post=post,
        patch=patch,
    )


def _create_meta_workflow_run_for_item_type(
    meta_workflow_run_creator_class: MetaWorkflowRunFromItem,
    item_identifier: str,
    meta_workflow_identifier: str,
    auth_key: JsonObject,
    post: bool = True,
    patch: bool = True,
) -> JsonObject:
    meta_workflow_run_creator = meta_workflow_run_creator_class(
        item_identifier, meta_workflow_identifier, auth_key
    )
    if post:
        if patch:
            meta_workflow_run_creator.post_and_patch()
        else:
            meta_workflow_run_creator.post_meta_workflow_run()
    return meta_workflow_run_creator.get_meta_workflow_run()


class MetaWorkflowRunCreationError(Exception):
    """Custom exception for error tracking."""


class MetaWorkflowRunFromItem:
    """Base class to hold common methods required to create/POST a
    MetaWorkflowRun and PATCH the Item used to create it.
    """

    # Embedding API Fields
    FIELDS_TO_GET = ["*", "project.uuid", "institution.uuid"]

    # Schema constants
    META_WORKFLOW_RUNS = "meta_workflow_runs"
    ASSOCIATED_META_WORKFLOW_RUN = "associated_meta_workflow_runs"
    PROJECT = "project"
    INSTITUTION = "institution"
    UUID = "uuid"
    META_WORKFLOW = "meta_workflow"
    FINAL_STATUS = "final_status"
    PENDING = "pending"
    WORKFLOW_RUNS = "workflow_runs"
    TITLE = "title"
    INPUT = "input"
    COMMON_FIELDS = "common_fields"
    FILES = "files"
    PROBAND_ONLY = "proband_only"
    INPUT_SAMPLES = "input_samples"
    ASSOCIATED_SAMPLE_PROCESSING = "associated_sample_processing"

    # Class constants
    META_WORKFLOW_RUN_ENDPOINT = "meta-workflow-runs"

    def __init__(self, input_item_identifier, meta_workflow_identifier, auth_key):
        """Initialize the object and set all attributes.

        :param meta_workflow_identifier: MetaWorkflow[portal] UUID,
            @id, or accession
        :type meta_workflow_identifier: str
        :param auth_key: Portal authorization key
        :type auth_key: dict
        :raises MetaWorkflowRunCreationError: If required item cannot
            be found on environment of authorization key
        """
        self.auth_key = auth_key
        self.input_item = make_embed_request(
            input_item_identifier,
            self.FIELDS_TO_GET,
            self.auth_key,
            single_item=True,
        )
        if not self.input_item:
            raise MetaWorkflowRunCreationError(
                "No Item found for given identifier: %s" % input_item_identifier
            )
        self.meta_workflow = self.get_item_properties(meta_workflow_identifier)
        if not self.meta_workflow:
            raise MetaWorkflowRunCreationError(
                "No MetaWorkflow found for given identifier: %s"
                % meta_workflow_identifier
            )
        self.proband_only = self.meta_workflow.get(self.PROBAND_ONLY, False)
        self.project = self.input_item.get(self.PROJECT)
        self.institution = self.input_item.get(self.INSTITUTION)
        self.input_item_uuid = self.input_item.get(self.UUID)
        self.existing_meta_workflow_runs = self.input_item.get(
            self.META_WORKFLOW_RUNS, []
        )
        self.meta_workflow_run_uuid = str(uuid.uuid4())
        self.meta_workflow_run = {}  # Overwrite in child classes

    def get_meta_workflow_run(self) -> JsonObject:
        return self.meta_workflow_run

    def create_workflow_runs(self, meta_workflow_run):
        """Create shards and update MetaWorkflowRun[json].

        :param meta_workflow_run: MetaWorkflowRun[json]
        :type meta_workflow_run: dict
        :raises MetaWorkflowRunCreationError: If input files to
            MetaWorkflowRun cannot be identified
        """
        reformatted_file_input = None
        reformatted_meta_workflow_run = MetaWorkflowRun(meta_workflow_run).to_json()
        reformatted_input = reformatted_meta_workflow_run[self.INPUT]
        for input_item in reformatted_input:
            input_files = input_item.get(self.FILES)
            if input_files is None:
                continue
            reformatted_file_input = input_files
            break
        if reformatted_file_input is None:
            raise MetaWorkflowRunCreationError(
                "No input files were provided for the MetaWorkflowRun: %s"
                % meta_workflow_run
            )
        run_with_workflows = MetaWorkflow(self.meta_workflow).write_run(
            reformatted_file_input
        )
        meta_workflow_run[self.WORKFLOW_RUNS] = run_with_workflows[self.WORKFLOW_RUNS]

    def get_item_properties(self, item_identifier):
        """Retrieve item from given environment without raising
        exception if not found.
        :param item_identifier: Item identifier on the portal
        :type item_identifier: str
        :return: Raw view of item if found
        :rtype: dict or None
        """
        try:
            result = ff_utils.get_metadata(
                item_identifier, key=self.auth_key, add_on="frame=raw"
            )
        except Exception:
            result = None
        return result

    def post_and_patch(self):
        """POST MetaWorkflowRun[json] and PATCH SampleProcessing[portal]
        to update list of its linked MetaWorkflowRun[portal].
        """
        self.post_meta_workflow_run()
        self.patch_input_item()

    def post_meta_workflow_run(self):
        """POST MetaWorkflowRun[json] to portal."""
        try:
            ff_utils.post_metadata(
                self.meta_workflow_run,
                self.META_WORKFLOW_RUN_ENDPOINT,
                key=self.auth_key,
            )
        except Exception as error_msg:
            raise MetaWorkflowRunCreationError(
                "MetaWorkflowRun not POSTed: \n%s" % str(error_msg)
            )

    def patch_input_item(self):
        """PATCH input item with link to the new MetaWorkflowRun[portal].

        We assume property to patch is self.META_WORKFLOW_RUNS, which
        is true for Sample and SampleProcessing.

        Method will fail unless MetaWorkflowRun[json] previously
        POSTed.
        """
        meta_workflow_runs = [
            identifier for identifier in self.existing_meta_workflow_runs
        ]
        meta_workflow_runs.append(self.meta_workflow_run_uuid)
        patch_body = {self.META_WORKFLOW_RUNS: meta_workflow_runs}
        try:
            ff_utils.patch_metadata(
                patch_body, obj_id=self.input_item_uuid, key=self.auth_key
            )
        except Exception as error_msg:
            raise MetaWorkflowRunCreationError(
                "Item could not be PATCHed: \n%s" % str(error_msg)
            )

    def _get_meta_workflow_run_title(self) -> str:
        meta_workflow_title = self.meta_workflow.get(self.TITLE)
        today = datetime.date.today().isoformat()
        return f"MetaWorkflowRun {meta_workflow_title} from {today}"

    def _get_meta_workflow_run_common_fields(self) -> JsonObject:
        return {
            self.PROJECT: self._get_project_uuid(),
            self.INSTITUTION: self._get_institution_uuid(),
            self.ASSOCIATED_META_WORKFLOW_RUN: [self.meta_workflow_run_uuid],
        }

    def _get_project_uuid(self) -> str:
        return self._get_uuid_or_raise_exception(self.project, self.PROJECT)

    def _get_institution_uuid(self) -> str:
        return self._get_uuid_or_raise_exception(self.institution, self.INSTITUTION)

    def _get_uuid_or_raise_exception(self, item: Any, item_type: str) -> str:
        if not isinstance(item, dict):
            raise TypeError(f"{item_type.capitalize()} was not embedded as an object")
        uuid = item.get(self.UUID)
        if uuid:
            return uuid
        raise ValueError(f"UUID was not found on the {item_type}")


################################################
#   MetaWorkflowRunFromSampleProcessing
################################################
class MetaWorkflowRunFromSampleProcessing(MetaWorkflowRunFromItem):
    """Class to create and POST|PATCH to portal a MetaWorkflowRun[json]
    from a SampleProcessing[portal] and a MetaWorkflow[portal].
    """

    # Embedding API fields
    FIELDS_TO_GET = [
        "samples.bam_sample_id",
        "samples.uuid",
        "samples.files.uuid",
        "samples.files.paired_end",
        "samples.files.file_format.file_format",
        "samples.files.related_files.relationship_type",
        "samples.files.related_files.file.uuid",
        "samples.files.related_files.file.file_format.file_format",
        "samples.files.related_files.file.paired_end",
        "samples.processed_files.uuid",
        "samples.processed_files.paired_end",
        "samples.processed_files.file_format.file_format",
        "files.uuid",
        "files.file_format.file_format",
        "files.variant_type",
    ] + MetaWorkflowRunFromItem.FIELDS_TO_GET

    def __init__(
        self,
        sample_processing_identifier,
        meta_workflow_identifier,
        auth_key,
        expect_family_structure=True,
    ):
        """Initialize the object and set all attributes.

        :param sample_processing_identifier: SampleProcessing[portal] UUID
            or @id
        :type sample_processing_identifier: str
        :param meta_workflow_identifier: MetaWorkflow[portal] UUID,
            @id, or accession
        :type meta_workflow_identifier: str
        :param auth_key: Portal authorization key
        :type auth_key: dict
        :param expect_family_structure: Whether a family structure is
            expected on the SampleProcessing[portal]
        :type expect_family_structure: bool
        :raises MetaWorkflowRunCreationError: If required item cannot
            be found on environment of authorization key
        """
        super().__init__(
            sample_processing_identifier, meta_workflow_identifier, auth_key
        )
        self.input_properties = InputPropertiesFromSampleProcessing(
            self.input_item,
            expect_family_structure=expect_family_structure,
            proband_only=self.proband_only,
        )
        self.meta_workflow_run_input = MetaWorkflowRunInput(
            self.meta_workflow, self.input_properties
        ).create_input()
        self.meta_workflow_run = self.create_meta_workflow_run()

    def create_meta_workflow_run(self):
        """Create MetaWorkflowRun[json] to later POST to portal.

        :return: MetaWorkflowRun[json]
        :rtype: dict
        """
        meta_workflow_run = {
            self.META_WORKFLOW: self.meta_workflow.get(self.UUID),
            self.INPUT: self.meta_workflow_run_input,
            self.TITLE: self._get_meta_workflow_run_title(),
            self.PROJECT: self._get_project_uuid(),
            self.INSTITUTION: self._get_institution_uuid(),
            self.INPUT_SAMPLES: self.input_properties.input_sample_uuids,
            self.ASSOCIATED_SAMPLE_PROCESSING: self.input_item_uuid,
            self.COMMON_FIELDS: self._get_meta_workflow_run_common_fields(),
            self.FINAL_STATUS: self.PENDING,
            self.WORKFLOW_RUNS: [],
            self.UUID: self.meta_workflow_run_uuid,
        }
        self.create_workflow_runs(meta_workflow_run)
        return meta_workflow_run


class MetaWorkflowRunFromSample(MetaWorkflowRunFromItem):
    """Class to create and POST/PATCH to portal a MetaWorkflowRun from
    a Sample and MetaWorkflow.
    """

    # Embedding API fields
    FIELDS_TO_GET = [
        "files.uuid",
        "files.paired_end",
        "files.file_format.file_format",
        "files.related_files.relationship_type",
        "files.related_files.file.uuid",
        "files.related_files.file.file_format.file_format",
        "files.related_files.file.paired_end",
        "processed_files.uuid",
        "processed_files.paired_end",
        "processed_files.file_format.file_format",
    ] + MetaWorkflowRunFromItem.FIELDS_TO_GET

    def __init__(self, sample_identifier, meta_workflow_identifier, auth_key):
        """Initialize the object and set all attributes.
        :param sample_identifier: Sample UUID or @id
        :type sample_identifier: str
        :param meta_workflow_identifier: MetaWorkflow[portal] UUID,
            @id, or accession
        :type meta_workflow_identifier: str
        :param auth_key: Portal authorization key
        :type auth_key: dict
        :raises MetaWorkflowRunCreationError: If required item cannot
            be found on environment of authorization key
        """
        super().__init__(sample_identifier, meta_workflow_identifier, auth_key)
        self.input_properties = InputPropertiesFromSample(self.input_item)
        self.meta_workflow_run_input = MetaWorkflowRunInput(
            self.meta_workflow, self.input_properties
        ).create_input()
        self.meta_workflow_run = self.create_meta_workflow_run()

    def create_meta_workflow_run(self):
        """Create MetaWorkflowRun[json] to later POST to portal.

        :return: MetaWorkflowRun[json]
        :rtype: dict
        """
        meta_workflow_run = {
            self.META_WORKFLOW: self.meta_workflow.get(self.UUID),
            self.INPUT: self.meta_workflow_run_input,
            self.TITLE: self._get_meta_workflow_run_title(),
            self.PROJECT: self._get_project_uuid(),
            self.INSTITUTION: self._get_institution_uuid(),
            self.INPUT_SAMPLES: self.input_properties.input_sample_uuids,
            self.COMMON_FIELDS: self._get_meta_workflow_run_common_fields(),
            self.FINAL_STATUS: self.PENDING,
            self.WORKFLOW_RUNS: [],
            self.UUID: self.meta_workflow_run_uuid,
        }
        self.create_workflow_runs(meta_workflow_run)
        return meta_workflow_run


def _get_sample_fields_to_embed(sample_prefix: str) -> List[str]:
    sample_fields_to_get = [
        "bam_sample_id",
        "uuid",
        "files.uuid",
        "files.file_format.file_format",
        "processed_files.uuid",
        "processed_files.file_format.file_format",
        "tissue_type",
    ]
    return [f"{sample_prefix}.{field}" for field in sample_fields_to_get]


class MetaWorkflowRunFromCohortAnalysis(MetaWorkflowRunFromItem):
    FIELDS_TO_GET = (
        _get_sample_fields_to_embed("case_samples")
        + _get_sample_fields_to_embed("control_samples")
        + MetaWorkflowRunFromItem.FIELDS_TO_GET
    )

    def __init__(
        self,
        cohort_analysis_identifier: str,
        meta_workflow_identifier: str,
        auth_key: JsonObject,
    ):
        """Initialize the object and set all attributes.
        :param cohort_analysis_identifier: cohort_analysis UUID or @id
        :param meta_workflow_identifier: MetaWorkflow[portal] UUID,
            @id, or accession
        :param auth_key: Portal authorization key
        :raises MetaWorkflowRunCreationError: If required item cannot
            be found on environment of authorization key
        """
        super().__init__(cohort_analysis_identifier, meta_workflow_identifier, auth_key)
        self.input_properties = InputPropertiesFromCohortAnalysis(self.input_item)
        self.meta_workflow_run_input = MetaWorkflowRunInput(
            self.meta_workflow, self.input_properties
        ).create_input()
        self.meta_workflow_run = self.create_meta_workflow_run()

    def create_meta_workflow_run(self) -> JsonObject:
        """Create MetaWorkflowRun[json] to later POST to portal.

        :return: MetaWorkflowRun[json]
        """
        meta_workflow_run = {
            self.META_WORKFLOW: self.meta_workflow.get(self.UUID),
            self.INPUT: self.meta_workflow_run_input,
            self.TITLE: self._get_meta_workflow_run_title(),
            self.PROJECT: self._get_project_uuid(),
            self.INSTITUTION: self._get_institution_uuid(),
            self.COMMON_FIELDS: self._get_meta_workflow_run_common_fields(),
            self.FINAL_STATUS: self.PENDING,
            self.WORKFLOW_RUNS: [],
            self.UUID: self.meta_workflow_run_uuid,
        }
        self.create_workflow_runs(meta_workflow_run)
        return meta_workflow_run


class MetaWorkflowRunFromSomaticAnalysis(MetaWorkflowRunFromItem):
    FIELDS_TO_GET = (
        [
            "processed_files.*",
            "processed_files.file_format.*",
            "individual.*",
        ]
        + _get_sample_fields_to_embed("samples")
        + MetaWorkflowRunFromItem.FIELDS_TO_GET
    )

    def __init__(
        self,
        somatic_analysis_identifier: str,
        meta_workflow_identifier: str,
        auth_key: JsonObject,
    ):
        """Initialize the object and set all attributes.
        :param somatic_analysis_identifier: somatic_analysis UUID or @id
        :param meta_workflow_identifier: MetaWorkflow[portal] UUID,
            @id, or accession
        :param auth_key: Portal authorization key
        :raises MetaWorkflowRunCreationError: If required item cannot
            be found on environment of authorization key
        """
        super().__init__(
            somatic_analysis_identifier, meta_workflow_identifier, auth_key
        )
        self.input_properties = InputPropertiesFromSomaticAnalysis(self.input_item)
        self.meta_workflow_run_input = MetaWorkflowRunInput(
            self.meta_workflow, self.input_properties
        ).create_input()
        self.meta_workflow_run = self.create_meta_workflow_run()

    def create_meta_workflow_run(self) -> JsonObject:
        """Create MetaWorkflowRun[json] to later POST to portal.

        :return: MetaWorkflowRun[json]
        """
        meta_workflow_run = {
            self.META_WORKFLOW: self.meta_workflow.get(self.UUID),
            self.INPUT: self.meta_workflow_run_input,
            self.TITLE: self._get_meta_workflow_run_title(),
            self.PROJECT: self._get_project_uuid(),
            self.INSTITUTION: self._get_institution_uuid(),
            self.COMMON_FIELDS: self._get_meta_workflow_run_common_fields(),
            self.FINAL_STATUS: self.PENDING,
            self.WORKFLOW_RUNS: [],
            self.UUID: self.meta_workflow_run_uuid,
        }
        self.create_workflow_runs(meta_workflow_run)
        return meta_workflow_run


################################################
#   MetaWorkflowRunInput
################################################
class MetaWorkflowRunInput:
    """Generate MetaWorkflowRun[json] input given MetaWorkflow[json] and
    input properties object with required input fields.
    """

    # Schema constants
    ARGUMENT_NAME = "argument_name"
    ARGUMENT_TYPE = "argument_type"
    VALUE = "value"
    VALUE_TYPE = "value_type"
    FILE = "file"
    FILES = "files"
    PARAMETER = "parameter"
    DIMENSION = "dimension"
    DIMENSIONALITY = "dimensionality"
    INPUT = "input"
    UUID = "uuid"

    def __init__(self, meta_workflow, input_properties):
        """Initialize the object and set all attributes.

        :param meta_workflow: MetaWorkflow[json]
        :type meta_workflow: dict
        :param input_properties: Object containing expected input
            parameters for MetaWorkflow[json]
        :type input_properties: object
        """
        self.meta_workflow = meta_workflow
        self.input_properties = input_properties

    def create_input(self):
        """Create MetaWorkflowRun[json] input based on arguments specified in
        MetaWorkflow[json].

        :return: MetaWorkflowRun[json] input
        :rtype: dict
        :raises MetaWorkflowRunCreationError: If input argument provided
            from MetaWorkflow could not be handled
        """
        result = []
        input_files_to_fetch = []
        input_parameters_to_fetch = []
        input_files = []
        input_parameters = []
        meta_workflow_input = self.meta_workflow.get(self.INPUT, [])
        for input_arg in meta_workflow_input:
            if self.FILES in input_arg or self.VALUE in input_arg:
                continue
            input_arg_name = input_arg.get(self.ARGUMENT_NAME)
            input_arg_type = input_arg.get(self.ARGUMENT_TYPE)
            if input_arg_type == self.FILE:
                input_arg_dimensions = input_arg.get(self.DIMENSIONALITY)
                input_files_to_fetch.append((input_arg_name, input_arg_dimensions))
            elif input_arg_type == self.PARAMETER:
                parameter_value_type = input_arg.get(self.VALUE_TYPE)
                input_parameters_to_fetch.append((input_arg_name, parameter_value_type))
            else:
                raise MetaWorkflowRunCreationError(
                    "Found an unexpected MetaWorkflow input argument type (%s) for"
                    " MetaWorkflow with uuid: %s"
                    % (input_arg_type, self.meta_workflow.get(self.UUID))
                )
        if input_parameters_to_fetch:
            input_parameters = self.fetch_parameters(input_parameters_to_fetch)
            result += input_parameters
        if input_files_to_fetch:
            input_files = self.fetch_files(input_files_to_fetch)
            result += input_files
        return result

    def fetch_files(self, files_to_fetch):
        """Create file inputs for MetaWorkflowRun[json].

        :param files_to_fetch: File input arguments from MetaWorkflow[json]
        :type files_to_fetch: list((str, int))
        :return: Structured file input for MetaWorkflowRun[json]
        :rtype: list(dict)
        :raises MetaWorkflowRunCreationError: If file input argument name
            from MetaWorkflow could not be found on the input properties
            class
        """
        result = []
        for file_parameter, input_dimensions in files_to_fetch:
            try:
                file_parameter_value = getattr(
                    self.input_properties, file_parameter.lower()
                )
            except AttributeError:
                raise MetaWorkflowRunCreationError(
                    "Could not find input parameter: %s" % file_parameter
                )
            formatted_file_value = self.format_file_input_value(
                file_parameter, file_parameter_value, input_dimensions
            )
            file_parameter_result = {
                self.ARGUMENT_NAME: file_parameter,
                self.ARGUMENT_TYPE: self.FILE,
                self.FILES: formatted_file_value,
            }
            result.append(file_parameter_result)
        return result

    def format_file_input_value(
        self,
        file_parameter: str,
        file_input_value: List[List[str]],
        input_dimensions: int,
    ) -> List[JsonObject]:
        """Create one structured file input for MetaWorkflowRun[json].

        File inputs arrive as lists of lists of UUIDs for now. For
        dimension 1, each sub-list should contain <= 1 UUID. For
        dimension 2, any number of UUIDs can be in each sub-list.

        To add more dimensions, will need to refactor inputs and this
        method.

        :param file_parameter: Name of file input argument
        :param file_input_value: File input values
        :param input_dimensions: The number of dimensions to use for
            the given file parameter
        :return: Structured file argument input
        :raises MetaWorkflowRunCreationError: If expected dimensions
            could not be handled or an input of dimension 1 has more
            than 1 entry per sample (i.e. is 2 dimensional)
        """
        result = []
        for input_idx, file_input in enumerate(file_input_value):
            if not file_input:
                raise MetaWorkflowRunCreationError(
                    f"Found an empty list of input files for parameter {file_parameter}"
                )
            if input_dimensions == 1:
                if len(file_input_value) > 1:
                    if len(file_input) > 1:
                        raise MetaWorkflowRunCreationError(
                            f"Found multiple input files when only 1 was expected for"
                            f" parameter {file_parameter}: {file_input}"
                        )
                    for file_idx, file_uuid in enumerate(file_input):
                        if not isinstance(file_uuid, str):
                            raise MetaWorkflowRunCreationError(
                                f"File input for parameter {file_parameter} was"
                                f" unexpected. Exected input file dimension"
                                f" {input_dimensions} but received the following:"
                                f" {file_uuid}."
                            )
                        dimension = str(input_idx)
                        formatted_file_result = {
                            self.FILE: file_uuid,
                            self.DIMENSION: dimension,
                        }
                        result.append(formatted_file_result)
                else:
                    for file_idx, file_uuid in enumerate(file_input):
                        if not isinstance(file_uuid, str):
                            raise MetaWorkflowRunCreationError(
                                f"File input for parameter {file_parameter} was"
                                f" unexpected. Exected input file dimension"
                                f" {input_dimensions} but received the following:"
                                f" {file_uuid}."
                            )
                        dimension = str(file_idx)
                        formatted_file_result = {
                            self.FILE: file_uuid,
                            self.DIMENSION: dimension,
                        }
                        result.append(formatted_file_result)
            elif input_dimensions == 2:
                for file_uuid_idx, file_uuid in enumerate(file_input):
                    if not isinstance(file_uuid, str):
                        raise MetaWorkflowRunCreationError(
                            f"File input for parameter {file_parameter} was unexpected."
                            f" Exected input file dimension {input_dimensions} but"
                            f" received the following: {file_input_value}."
                        )
                    dimension = f"{input_idx},{file_uuid_idx}"
                    formatted_file_result = {
                        self.FILE: file_uuid,
                        self.DIMENSION: dimension,
                    }
                    result.append(formatted_file_result)
            else:
                raise MetaWorkflowRunCreationError(
                    f"Received an unexpected dimension number for parameter"
                    f" {file_parameter}: {input_dimensions}"
                )
        return result

    def fetch_parameters(self, parameters_to_fetch):
        """Create non-file parameters for MetaWorkflowRun[json].

        :param parameters_to_fetch: Non-file input parameters from
            MetaWorkflow[json]
        :type parameters_to_fetch: list((str, str))
        :return: Structured non-file input
        :rtype: list(dict)
        :raises MetaWorkflowRunCreationError: If given parameter could
            not be found on the input properties class
        """
        result = []
        for parameter, value_type in parameters_to_fetch:
            try:
                parameter_value = getattr(self.input_properties, parameter.lower())
            except AttributeError:
                raise MetaWorkflowRunCreationError(
                    "Could not find input parameter: %s" % parameter
                )
            parameter_value = self.cast_parameter_value(parameter_value)
            parameter_result = {
                self.ARGUMENT_NAME: parameter,
                self.ARGUMENT_TYPE: self.PARAMETER,
                self.VALUE: parameter_value,
                self.VALUE_TYPE: value_type,
            }
            result.append(parameter_result)
        return result

    def cast_parameter_value(self, parameter_value):
        """Cast parameter value in expected format based on value
        type.

        :param parameter_value: Value for a given input parameter
        :type parameter_value: object
        :return: Possibly JSON-formatted string representation of the
            value
        :rtype: str
        """
        if isinstance(parameter_value, list) or isinstance(parameter_value, dict):
            result = json.dumps(parameter_value)
        else:
            result = str(parameter_value)
        return result


def get_files_for_file_formats(file_items, file_formats, requirements=None):
    """Gather file UUIDs for File items matching the file format
    and meeting the requirements.

    :param file_items: The files to sort through
    :type file_items: list(dict)
    :param file_formats: Accepted formats of files to get
    :type file_formats: list or set
    :param requirements: Requirements a file must meet in order to
        be acceptable, as key, value pairs of property names, lists
        of acceptable property values
    :type requirements: dict
    :returns: File UUIDs of files meeting file format and
        requirements
    :rtype: list(str)
    """
    result = []
    for file_item in file_items:
        file_item_format = file_item.get(FILE_FORMAT, {}).get(FILE_FORMAT)
        if file_item_format in file_formats:
            if requirements:
                if not all(
                    file_item.get(key) in accepted_values
                    for key, accepted_values in requirements.items()
                ):
                    continue
            file_uuid = file_item.get(UUID)
            result.append(file_uuid)
    return result


################################################
#   InputPropertiesFromSampleProcessing
################################################
class InputPropertiesFromSampleProcessing:
    """Class for accessing MetaWorkflowRun[json] input arguments from a
    SampleProcessing[json].
    """

    # Schema constants
    UUID = "uuid"
    BAM_SAMPLE_ID = "bam_sample_id"
    SAMPLES_PEDIGREE = "samples_pedigree"
    SAMPLES = "samples"
    RELATIONSHIP = "relationship"
    PROBAND = "proband"
    MOTHER = "mother"
    FATHER = "father"
    INDIVIDUAL = "individual"
    PARENTS = "parents"
    SAMPLE_NAME = "sample_name"
    SEX = "sex"
    FILES = "files"
    FILE_FORMAT = "file_format"
    VARIANT_TYPE = "variant_type"
    SNV_VARIANT_TYPE = "SNV"
    SV_VARIANT_TYPE = "SV"

    # Class constants
    GENDER = "gender"
    VCF_GZ_FORMAT = "vcf_gz"
    VCF_FORMAT = "vcf"
    VCF_FORMATS = set([VCF_GZ_FORMAT, VCF_FORMAT])

    def __init__(
        self,
        sample_processing,
        expect_family_structure=True,
        proband_only=False,
    ):
        """Initialize the object and set attributes.

        :param sample_processing: SampleProcessing[json]
        :type sample_processing: dict
        :param expect_family_structure: Whether a family structure is
            expected on the SampleProcessing, which influences sample
            sorting and pedigree expectations
        :type expect_family_structure: bool
        :param proband_only: Whether inputs should only use proband
            samples. Only applicable within family structure.
        :type proband_only: bool
        """
        self.expect_family_structure = expect_family_structure
        self.sample_processing = sample_processing
        self.samples = sample_processing.get(self.SAMPLES)
        if not self.samples:
            raise MetaWorkflowRunCreationError(
                "No Samples found on SampleProcessing: %s" % self.sample_processing
            )
        self.samples_pedigree = sample_processing.get(self.SAMPLES_PEDIGREE, [])
        if expect_family_structure:
            self.clean_and_sort_samples_and_pedigree()
        if proband_only:
            self.remove_non_proband_samples()
        self.sample_inputs = [
            InputPropertiesFromSample(sample) for sample in self.samples
        ]

    def clean_and_sort_samples_and_pedigree(self):
        """Sort Samples and pedigree and remove parents from pedigree
        if not included in Samples.

        Sorting order will be proband, then mother, then father, as
        applicable.

        :raises MetaWorkflowRunCreationError: If pedigree not found,
            samples and pegigree of different lengths, or required
            properties not found
        """
        proband_name = None
        mother_name = None
        father_name = None
        if not self.samples_pedigree:
            raise MetaWorkflowRunCreationError(
                "No samples_pedigree found on SampleProcessing: %s"
                % self.sample_processing
            )
        if len(self.samples) != len(self.samples_pedigree):
            raise MetaWorkflowRunCreationError(
                "Number of Samples did not match number of entries in samples_pedigree"
                " on SampleProcessing: %s" % self.sample_processing
            )
        all_individuals = [
            sample.get(self.INDIVIDUAL)
            for sample in self.samples_pedigree
            if sample.get(self.INDIVIDUAL)
        ]
        bam_sample_ids = [
            sample.get(self.BAM_SAMPLE_ID)
            for sample in self.samples
            if sample.get(self.BAM_SAMPLE_ID)
        ]
        for pedigree_sample in self.samples_pedigree:
            parents = pedigree_sample.get(self.PARENTS, [])
            if parents:  # Remove parents that aren't in samples_pedigree
                missing_parents = [
                    parent for parent in parents if parent not in all_individuals
                ]
                for missing_parent in missing_parents:
                    parents.remove(missing_parent)
            sample_name = pedigree_sample.get(self.SAMPLE_NAME)
            if sample_name is None:
                raise MetaWorkflowRunCreationError(
                    "No sample name given for sample in pedigree: %s" % pedigree_sample
                )
            elif sample_name not in bam_sample_ids:
                raise MetaWorkflowRunCreationError(
                    "Sample in pedigree not found on SampleProcessing: %s" % sample_name
                )
            sex = pedigree_sample.get(self.SEX)
            if sex is None:
                raise MetaWorkflowRunCreationError(
                    "No sex given for sample in pedigree: %s" % pedigree_sample
                )
            relationship = pedigree_sample.get(self.RELATIONSHIP)
            if relationship == self.PROBAND:
                proband_name = sample_name
            elif relationship == self.MOTHER:
                mother_name = sample_name
            elif relationship == self.FATHER:
                father_name = sample_name
        if proband_name is None:
            raise MetaWorkflowRunCreationError(
                "No proband found within the pedigree: %s" % self.samples_pedigree
            )
        self.sort_by_sample_name(
            self.samples_pedigree,
            self.SAMPLE_NAME,
            proband_name,
            mother=mother_name,
            father=father_name,
        )
        self.sort_by_sample_name(
            self.samples,
            self.BAM_SAMPLE_ID,
            proband_name,
            mother=mother_name,
            father=father_name,
        )

    def remove_non_proband_samples(self):
        """Remove all non-proband members from samples and pedigree."""
        if self.expect_family_structure:  # Already sorted to proband-first
            del self.samples_pedigree[1:]
            del self.samples[1:]
        else:
            proband_only_samples = []
            proband_only_samples_pedigree = []
            proband_sample_names = self.probands
            for sample in self.samples:
                sample_name = sample.get(self.BAM_SAMPLE_ID)
                if sample_name in proband_sample_names:
                    proband_only_samples.append(sample)
            for pedigree_sample in self.samples_pedigree:
                sample_name = pedigree_sample.get(self.SAMPLE_NAME)
                if sample_name in proband_sample_names:
                    proband_only_samples_pedigree.append(pedigree_sample)
            self.samples = proband_only_samples
            self.samples_pedigree = proband_only_samples_pedigree

    def sort_by_sample_name(
        self, items_to_sort, sample_name_key, proband, mother=None, father=None
    ):
        """Sort items to be proband, mother, father, then other family
        members, as applicable.

        :param items_to_sort: Items to sort
        :type items_to_sort: list(dict)
        :param sample_name_key: Key of the item dict corresponding to
            sample name of the item
        :type sample_name_key: str
        :param proband: Proband sample name
        :type proband: str
        :param mother: Mother sample name
        :type mother: str or None
        :param father: Father sample name
        :type father: str or None
        """
        sorted_items = []
        other_idx = []
        proband_idx = None
        mother_idx = None
        father_idx = None
        for idx, item in enumerate(items_to_sort):
            sample_name = item.get(sample_name_key)
            if sample_name == proband:
                proband_idx = idx
            elif mother and sample_name == mother:
                mother_idx = idx
            elif father and sample_name == father:
                father_idx = idx
            else:
                other_idx.append(idx)
        if proband_idx is not None:
            sorted_items.append(items_to_sort[proband_idx])
        if mother_idx is not None:
            sorted_items.append(items_to_sort[mother_idx])
        if father_idx is not None:
            sorted_items.append(items_to_sort[father_idx])
        for idx in other_idx:
            sorted_items.append(items_to_sort[idx])
        items_to_sort.clear()
        items_to_sort.extend(sorted_items)

    def get_property_from_samples(self, sample_property):
        """Retrieve input property from samples on this
        SampleProcessing.

        Input property names must align here and on corresponding
        class for Samples.

        :param sample_property: The MWF input argument to get
        :type sample_property: str
        :return: Property retrieved from all Samples on item
        :rtype: list
        """
        result = []
        for sample_input in self.sample_inputs:
            result += getattr(sample_input, sample_property)
        return result

    def get_submitted_vcf_files(self, requirements=None):
        """Collect submitted VCFs meeting given requirements.

        :param requirements: Property requirements for the VCF files to
            meet
        :type requirements: dict
        :returns: UUIDs of found VCF files
        :rtype: list
        :raises MetaWorkflowRunCreationError: If no VCF files meeting
            requirements found
        """
        submitted_files = self.sample_processing.get(self.FILES, [])
        result = get_files_for_file_formats(
            submitted_files, self.VCF_FORMATS, requirements=requirements
        )
        if not result:
            raise MetaWorkflowRunCreationError(
                f"No file with acceptable VCF file format meeting requirements found on"
                f" SampleProcessing: {self.sample_processing}"
            )
        return [result]

    # SampleProcessing-specific properties
    @property
    def pedigree(self):
        """Sorted pedigree input."""
        result = []
        for pedigree_sample in self.samples_pedigree:
            result.append(
                {
                    self.PARENTS: pedigree_sample.get(self.PARENTS, []),
                    self.INDIVIDUAL: pedigree_sample.get(self.INDIVIDUAL, ""),
                    self.SAMPLE_NAME: pedigree_sample.get(self.SAMPLE_NAME),
                    # May want to switch gender key to sex below
                    self.GENDER: pedigree_sample.get(self.SEX),
                }
            )
        return result

    @property
    def sample_name_proband(self):
        """Proband Sample name input."""
        return [self.sample_names[0]]  # Already sorted to proband-first

    @property
    def bamsnap_titles(self):
        """Sorted BAMSnap name input."""
        result = []
        for sample_pedigree in self.samples_pedigree:
            sample_name = sample_pedigree.get(self.SAMPLE_NAME)
            sample_relationship = sample_pedigree.get(self.RELATIONSHIP, "")
            result.append("%s (%s)" % (sample_name, sample_relationship))
        return result

    @property
    def family_size(self):
        """Family size input."""
        return len(self.sample_names)

    @property
    def probands(self):
        """Proband list for jointly calling multiple g.vcf files"""
        probands = []
        if not self.samples_pedigree:
            raise MetaWorkflowRunCreationError(
                "Cannot create probands list because there is no samples pedigree."
            )
        for pedigree_sample in self.samples_pedigree:
            relationship = pedigree_sample.get(self.RELATIONSHIP)
            sample_name = pedigree_sample.get(self.SAMPLE_NAME)
            if (relationship is None) or (sample_name is None):
                raise MetaWorkflowRunCreationError(
                    "Pedigree information incomplete. Relationship and sample name"
                    " are required."
                )
            if relationship == self.PROBAND:
                probands.append(sample_name)
        return probands

    @property
    def input_vcfs(self):
        """VCFs submitted to the SampleProcessing."""
        return self.get_submitted_vcf_files()

    @property
    def input_snv_vcfs(self):
        """SNV VCFs submitted to the SampleProcessing."""
        requirements = {self.VARIANT_TYPE: [self.SNV_VARIANT_TYPE]}
        return self.get_submitted_vcf_files(requirements=requirements)

    @property
    def input_sv_vcfs(self):
        """SV VCFs submitted to the SampleProcessing."""
        requirements = {self.VARIANT_TYPE: [self.SV_VARIANT_TYPE]}
        return self.get_submitted_vcf_files(requirements=requirements)

    # Properties from Samples
    @property
    def sample_names(self):
        """Sorted Sample name input."""
        return self.get_property_from_samples("sample_names")

    @property
    def input_sample_uuids(self):
        """Sorted Sample UUID input."""
        return self.get_property_from_samples("input_sample_uuids")

    @property
    def input_crams(self):
        """CRAM file input."""
        return self.get_property_from_samples("input_crams")

    @property
    def input_gvcfs(self):
        """gVCF file input."""
        return self.get_property_from_samples("input_gvcfs")

    @property
    def fastqs_r1(self):
        """FASTQ paired-end 1 file input."""
        return self.get_property_from_samples("fastqs_r1")

    @property
    def fastqs_r2(self):
        """FASTQ paired-end 2 file input."""
        return self.get_property_from_samples("fastqs_r2")

    @property
    def input_bams(self):
        """BAM file input."""
        return self.get_property_from_samples("input_bams")

    @property
    def rcktar_file_names(self):
        """Sorted names for created RckTar files input."""
        return self.get_property_from_samples("rcktar_file_names")


class InputPropertiesFromSample:
    """Class for accessing MetaWorkflowRun input arguments from a
    Sample item.
    """

    # Schema constants
    BAM_SAMPLE_ID = "bam_sample_id"
    FILE = "file"
    FILE_FORMAT = "file_format"
    FILES = "files"
    INDIVIDUAL = "individual"
    PAIRED_END = "paired_end"
    PAIRED_END_1 = "1"
    PAIRED_END_2 = "2"
    PAIRED_WITH = "paired with"
    PROCESSED_FILES = "processed_files"
    RELATED_FILES = "related_files"
    RELATIONSHIP_TYPE = "relationship_type"
    SEX = "sex"
    UUID = "uuid"

    # File formats
    FASTQ_FORMAT = "fastq"
    CRAM_FORMAT = "cram"
    BAM_FORMAT = "bam"
    GVCF_FORMAT = "gvcf_gz"

    # Class constants
    RCKTAR_FILE_ENDING = ".rck.gz"

    def __init__(self, sample):
        """Initialize the object and set attributes.

        :param sample: Sample metadata
        :type sample: dict
        """
        self.sample = sample

    def get_processed_files_for_file_format(self, file_format, requirements=None):
        """Get all files matching given file format on given sample
        that meet the given requirements.

        :param file_format: Format of files to get
        :type file_format: str
        :param requirements: Requirements a file must meet in order to
            be acceptable, as key, value pairs of property names, lists
            of acceptable property values
        :type requirements: dict
        :returns: Processed file UUIDs of files meeting file format and
            requirements
        :rtype: list(str)
        """
        result = []
        processed_files = self.sample.get(self.PROCESSED_FILES, [])
        result = get_files_for_file_formats(
            processed_files, [file_format], requirements=requirements
        )
        if not result:
            raise MetaWorkflowRunCreationError(
                "No file with format %s meeting requirements %s found on Sample: %s"
                % (file_format, requirements, self.sample)
            )
        return result

    def get_submitted_files_for_file_format(self, file_format, requirements=None):
        """Get all submitted file UUIDs matching file format and meeting
        requirements.

        :param file_format: Format of files to get
        :type file_format: str
        :param requirements: Requirements a file must meet in order to
            be acceptable, as key, value pairs of property names, lists
            of acceptable property values
        :type requirements: dict
        :returns: Submitted file UUIDs of files meeting file format and
            requirements
        :rtype: list(str)
        """
        submitted_files = self.sample.get(self.FILES, [])
        result = get_files_for_file_formats(
            submitted_files, [file_format], requirements=requirements
        )
        if not result:
            raise MetaWorkflowRunCreationError(
                "No file with format %s meeting requirements %s found on Sample: %s"
                % (file_format, requirements, self.sample)
            )
        return result

    def get_fastqs_for_paired_end(self, paired_end):
        """Get FASTQ files on Sample for given paired end.

        Searches for files first under Sample.files and then, if no
        matches found there, under Sample.processed_files.

        For non-ProcessedFile FASTQ files, enforce more stringent
        related file checks. For ProcessedFile FASTQs, related file
        property generally absent.

        :param paired_end: Desired paired end for FASTQs
        :type paired_end: str
        :return: FASTQ file UUIDs of matching paired end
        :rtype: list(str)
        :raises MetaWorkflowRunCreationError: If FASTQ file without
            related_files property found or no FASTQ files of matching
            paired end found
        """
        paired_end_fastqs = []
        fastq_files = self.sample.get(self.FILES, [])
        for fastq_file in fastq_files:
            file_uuid = fastq_file.get(self.UUID)
            if file_uuid in paired_end_fastqs:  # May have come as related file
                continue
            file_format = fastq_file.get(self.FILE_FORMAT, {}).get(self.FILE_FORMAT)
            if file_format != self.FASTQ_FORMAT:
                continue
            related_files = fastq_file.get(self.RELATED_FILES)
            if related_files is None:
                raise MetaWorkflowRunCreationError(
                    "Sample contains an uploaded FASTQ file without a related file: %s"
                    % self.sample
                )
            file_paired_end = fastq_file.get(self.PAIRED_END)
            if file_paired_end == paired_end:
                paired_end_fastqs.append(file_uuid)
                continue
            else:
                related_paired_end = False
                for related_file in related_files:
                    file_relation = related_file.get(self.RELATIONSHIP_TYPE)
                    if file_relation != self.PAIRED_WITH:
                        continue
                    file_properties = related_file.get(self.FILE, {})
                    related_uuid = file_properties.get(self.UUID)
                    related_paired_end = file_properties.get(self.PAIRED_END)
                    if related_paired_end == paired_end:
                        related_paired_end = True
                        if related_uuid not in paired_end_fastqs:
                            paired_end_fastqs.append(related_uuid)
                        break
                if related_paired_end is False:
                    raise MetaWorkflowRunCreationError(
                        "Found an uploaded FASTQ file without matching paired end (%s)"
                        " on the file or any of its related files: %s."
                        % (paired_end, fastq_file)
                    )

        if not paired_end_fastqs:  # May have come from CRAM conversion
            requirements = {self.PAIRED_END: [paired_end]}
            paired_end_fastqs = self.get_processed_files_for_file_format(
                self.FASTQ_FORMAT, requirements=requirements
            )
        return paired_end_fastqs

    @property
    def input_crams(self):
        """CRAM file input."""
        return [self.get_submitted_files_for_file_format(self.CRAM_FORMAT)]

    @property
    def input_gvcfs(self):
        """gVCF file input."""
        gvcfs = self.get_processed_files_for_file_format(self.GVCF_FORMAT)
        return [keep_last_item(gvcfs)]

    @property
    def fastqs_r1(self):
        """FASTQ paired-end 1 file input."""
        return [self.get_fastqs_for_paired_end(self.PAIRED_END_1)]

    @property
    def fastqs_r2(self):
        """FASTQ paired-end 2 file input."""
        return [self.get_fastqs_for_paired_end(self.PAIRED_END_2)]

    @property
    def fastqs(self):
        """User-uploaded FASTQ files, regardless of paired end."""
        result = []
        fastq_files = self.sample.get(self.FILES, [])
        for fastq_file in fastq_files:
            file_uuid = fastq_file.get(self.UUID)
            file_format = fastq_file.get(self.FILE_FORMAT, {}).get(self.FILE_FORMAT)
            if file_format == self.FASTQ_FORMAT:
                result.append(file_uuid)
        return [result]

    @property
    def input_bams(self):
        """BAM file input."""
        bams = self.get_processed_files_for_file_format(self.BAM_FORMAT)
        return [keep_last_item(bams)]

    @property
    def rcktar_file_names(self):
        """Name for created RckTar files input."""
        return [
            sample_name + self.RCKTAR_FILE_ENDING for sample_name in self.sample_names
        ]

    @property
    def sample_names(self):
        """Sample name from BAM"""
        return [self.sample[self.BAM_SAMPLE_ID]]

    @property
    def input_sample_uuids(self):
        """Sample UUID"""
        return [self.sample[self.UUID]]

    @property
    def individual(self) -> JsonObject:
        return self.sample.get(self.INDIVIDUAL)

    @property
    def sex(self) -> str:
        return self.individual.get(self.SEX)


class InputPropertiesFromCohortAnalysis:

    # Schema constants
    CASE_SAMPLES = "case_samples"
    CONTROL_SAMPLES = "control_samples"

    def __init__(self, cohort_analysis: JsonObject) -> None:
        """Initialize the object and set attributes.

        :param cohort_analysis: CohortAnalysis metadata
        :type cohort_analysis: dict
        """
        self.cohort_analysis = cohort_analysis
        case_samples = cohort_analysis.get(self.CASE_SAMPLES, [])
        self.case_sample_inputs = [
            InputPropertiesFromSample(sample) for sample in case_samples
        ]
        control_samples = cohort_analysis.get(self.CONTROL_SAMPLES, [])
        self.control_sample_inputs = [
            InputPropertiesFromSample(sample) for sample in control_samples
        ]

    def _get_input_property_from_all_samples(
        self, input_property_name: str
    ) -> List[Any]:
        result = []
        result += self._get_input_property_from_case_samples(input_property_name)
        result += self._get_input_property_from_control_samples(input_property_name)
        return result

    def _get_input_property_from_case_samples(
        self, input_property_name: str
    ) -> List[Any]:
        result = []
        for case_sample_input in self.case_sample_inputs:
            result += getattr(case_sample_input, input_property_name)
        return result

    def _get_input_property_from_control_samples(
        self, input_property_name: str
    ) -> List[Any]:
        result = []
        for control_sample_input in self.control_sample_inputs:
            result += getattr(control_sample_input, input_property_name)
        return result

    @property
    def input_gvcfs(self) -> List[List[str]]:
        return self._get_input_property_from_all_samples("input_gvcfs")


class InputPropertiesFromSomaticAnalysis:

    # Schema constants
    FILE_TYPE = "file_type"
    INDIVIDUAL = "individual"
    PROCESSED_FILES = "processed_files"
    SAMPLES = "samples"
    SEX = "sex"
    TISSUE_TYPE = "tissue_type"
    TISSUE_TYPE_NORMAL = "Normal"
    TISSUE_TYPE_TUMOR = "Tumor"

    TNSCOPE_FILE_TYPE = "TNscope VCF"
    FILE_FORMAT_VCF_GZ = "vcf_gz"

    def __init__(self, somatic_analysis: JsonObject) -> None:
        """Initialize the object and set attributes.

        :param sample_analysis: SampleAnalysis metadata
        :type sample_analysis: dict
        """
        self.somatic_analysis = somatic_analysis
        self._validate()

    def _validate(self) -> None:
        if len(self.samples) != len(self.tumor_samples) + len(self.normal_samples):
            raise MetaWorkflowRunCreationError(
                "Not every sample labelled as tumor or normal"
            )

    def _get_sample_type(self, sample: JsonObject) -> str:
        return sample.get(self.TISSUE_TYPE, "")

    def _is_normal_sample(self, sample: JsonObject) -> bool:
        return self._get_sample_type(sample) == self.TISSUE_TYPE_NORMAL

    def _is_tumor_sample(self, sample: JsonObject) -> bool:
        return self._get_sample_type(sample) == self.TISSUE_TYPE_TUMOR

    def _get_input_properties_from_sample_inputs(
        self, sample_inputs: List[InputPropertiesFromSample], input_property_name: str
    ) -> List[Any]:
        result = []
        for sample_input in sample_inputs:
            result += getattr(sample_input, input_property_name)
        return result

    def _get_input_properties_from_all_samples(
        self, input_property_name: str
    ) -> List[Any]:
        return self._get_input_properties_from_sample_inputs(
            self.all_sample_inputs, input_property_name
        )

    def _get_input_properties_from_tumor_samples(
        self, input_property_name: str
    ) -> List[Any]:
        return self._get_input_properties_from_sample_inputs(
            self.tumor_sample_inputs, input_property_name
        )

    def _get_input_properties_from_normal_samples(
        self, input_property_name: str
    ) -> List[Any]:
        return self._get_input_properties_from_sample_inputs(
            self.normal_sample_inputs, input_property_name
        )

    @property
    def individual(self) -> JsonObject:
        return self.somatic_analysis.get(self.INDIVIDUAL, {})

    @property
    def samples(self) -> List[JsonObject]:
        return self.somatic_analysis.get(self.SAMPLES, [])

    @property
    def tumor_samples(self) -> List[JsonObject]:
        return [sample for sample in self.samples if self._is_tumor_sample(sample)]

    @property
    def normal_samples(self) -> List[JsonObject]:
        return [sample for sample in self.samples if self._is_normal_sample(sample)]

    @property
    @lru_cache()
    def tumor_sample_inputs(self) -> List[InputPropertiesFromSample]:
        return [InputPropertiesFromSample(sample) for sample in self.tumor_samples]

    @property
    @lru_cache()
    def normal_sample_inputs(self) -> List[InputPropertiesFromSample]:
        return [InputPropertiesFromSample(sample) for sample in self.normal_samples]

    @property
    @lru_cache()
    def all_sample_inputs(self) -> List[InputPropertiesFromSample]:
        return self.tumor_sample_inputs + self.normal_sample_inputs

    @property
    def input_normal_bam(self) -> List[List[str]]:
        normal_bams = self._get_input_properties_from_normal_samples("input_bams")
        if len(normal_bams) != 1:
            raise MetaWorkflowRunCreationError
        return normal_bams

    @property
    def input_tumor_bam(self) -> List[List[str]]:
        tumor_bams = self._get_input_properties_from_tumor_samples("input_bams")
        if len(tumor_bams) != 1:
            raise MetaWorkflowRunCreationError
        return tumor_bams

    @property
    def sex(self) -> str:
        return self.individual.get(self.SEX, "")

    @property
    def tumor_sample_name(self) -> str:
        tumor_sample_names = self._get_input_properties_from_tumor_samples(
            "sample_names"
        )
        unique_sample_names = [name for name in set(tumor_sample_names) if name]
        if len(unique_sample_names) != 1:
            raise MetaWorkflowRunCreationError(
                "Could not identify only 1 tumor sample name"
            )
        return unique_sample_names[0]

    @property
    def normal_sample_name(self) -> str:
        normal_sample_names = self._get_input_properties_from_normal_samples(
            "sample_names"
        )
        unique_sample_names = [name for name in set(normal_sample_names) if name]
        if len(unique_sample_names) != 1:
            raise MetaWorkflowRunCreationError(
                "Could not identify only 1 normal sample name"
            )
        return unique_sample_names[0]

    @property
    def input_somatic_vcf(self) -> List[List[str]]:
        requirements = {self.FILE_TYPE: self.TNSCOPE_FILE_TYPE}
        processed_files = self.somatic_analysis.get(self.PROCESSED_FILES, [])
        result = get_files_for_file_formats(
            processed_files, self.FILE_FORMAT_VCF_GZ, requirements=requirements
        )
        if not result:
            raise MetaWorkflowRunCreationError(
                f"No file with acceptable VCF file format meeting requirements found on"
                f" SomaticAnalysis: {self.somatic_analysis}"
            )
        return [result]



#!/usr/bin/env python3

################################################
#
#   Function to import from old
#       MetaWorkflowRun[portal]
#
################################################
import json

from dcicutils import ff_utils

from magma.metawflrun import MetaWorkflowRun
from magma_ff.create_metawfr import MetaWorkflowRunFromSampleProcessing
from magma_ff.runupdate import RunUpdate
from magma_ff.utils import make_embed_request


################################################
#   Functions
################################################
def import_metawfr(
    meta_workflow_uuid,
    metawfr_uuid,
    sample_processing_uuid,
    steps_name,
    ff_key,
    post=False,
    patch=False,
    verbose=False,
    expect_family_structure=True,
):
    """Create new MetaWorkflowRun[json] from existing MetaWorkflowRun[portal].
    Import specified steps. If post, POST the MetaWorkflowRun[json] to portal
    and PATCH SampleProcessing[portal].

    :param meta_workflow_uuid: MetaWorkflow[portal] UUID
    :type meta_workflow_uuid: str
    :param metawfr_uuid: Existing MetaWorkflowRun[portal] UUID
    :type metawfr_uuid: str
    :param sample_processing_uuid: SampleProcessing[portal] UUID
    :type sample_processing_uuid: str
    :param steps_name: Names for the steps to import from existing
        MetaWorkflowRun[portal]
    :type steps_name: list(str)
    :param ff_key: Portal authorization key
    :type ff_key: dict
    :param post: Whether to POST the new MetaWorkflowRun[json]
    :type post: bool
    :param patch: Whether to PATCH new MetaWorkflowRun[json] to
        SampleProcessing[portal]
    :type patch: bool
    :param verbose: Whether to print the POST response
    :type verbose: bool
    :param expect_family_structure: Whether a family structure is
        expected for the SampleProcessing
    :type expect_family_structure: bool
    :return: New MetaWorkflowRun[json]
    :rtype: dict
    """
    run_json_to_import = ff_utils.get_metadata(
        metawfr_uuid, add_on="frame=raw", key=ff_key
    )
    run_obj_to_import = MetaWorkflowRun(run_json_to_import)
    meta_workflow_run_creator = MetaWorkflowRunFromSampleProcessing(
        sample_processing_uuid,
        meta_workflow_uuid,
        ff_key,
        expect_family_structure=expect_family_structure,
    )
    new_meta_workflow_run = meta_workflow_run_creator.meta_workflow_run
    run_obj = MetaWorkflowRun(new_meta_workflow_run)
    runupd_obj = RunUpdate(run_obj)
    updated_meta_workflow_run = runupd_obj.import_steps(run_obj_to_import, steps_name)
    # A little hacky, but replace the MWFR on the creator class to access POST and
    # PATCH methods
    meta_workflow_run_creator.meta_workflow_run = updated_meta_workflow_run
    if post:
        post_response = meta_workflow_run_creator.post_meta_workflow_run()
        if verbose:
            print(post_response)
        update_success, update_errors = add_new_meta_workflow_run_uuid_to_copied_items(
            updated_meta_workflow_run, ff_key
        )
        if verbose:
            print(
                "Updated %s copied MWFR output items with new MWFR UUID"
                % len(update_success)
            )
        if update_errors:  # Not raising Exception here so SampleProcessing PATCH occurs
            print(
                "WARNING: %s items were not updated with new MWFR UUID"
                % len(update_errors)
            )
            print("UUIDs and errors:")
            print(json.dumps(update_errors, indent=4))
    if patch:
        patch_response = meta_workflow_run_creator.patch_sample_processing()
        if verbose:
            print(patch_response)
    return updated_meta_workflow_run


def add_new_meta_workflow_run_uuid_to_copied_items(meta_workflow_run, auth_key):
    """For all output items copied, add new MWFR to
    Item.associated_meta_workflow_runs.

    :param meta_workflow_run: MetaWorkflowRun properties
    :type meta_workflow_run: dict
    :param auth_key: Portal authorization key
    :type auth_key: dict
    :return: Successful PATCHes and items with errors on PATCH
    :rtype: tuple(list, dict)
    """
    success = []
    errors = {}
    patch_items_and_bodies = {}
    workflow_run_uuids = []
    workflow_run_embed_fields = [
        "uuid",
        "associated_meta_workflow_runs",
        "output_files.value.uuid",
        "output_files.value.associated_meta_workflow_runs",
        "output_files.value_qc.uuid",
        "output_files.value_qc.associated_meta_workflow_runs",
    ]
    meta_workflow_run_uuid = meta_workflow_run.get("uuid")
    workflow_runs = meta_workflow_run.get("workflow_runs", [])
    for item in workflow_runs:
        workflow_run_uuid = item.get("workflow_run")
        if workflow_run_uuid:
            workflow_run_uuids.append(workflow_run_uuid)
    embed_result = make_embed_request(
        workflow_run_uuids, workflow_run_embed_fields, auth_key
    )
    for result in embed_result:
        add_to_patch(patch_items_and_bodies, result, meta_workflow_run_uuid)
        output_files = result.get("output_files", [])
        for output_file in output_files:
            file_item = output_file.get("value")
            add_to_patch(patch_items_and_bodies, file_item, meta_workflow_run_uuid)
            qc_file_item = output_file.get("value_qc")
            add_to_patch(patch_items_and_bodies, qc_file_item, meta_workflow_run_uuid)
    for item_to_patch, patch_body in patch_items_and_bodies.items():
        try:
            ff_utils.patch_metadata(patch_body, obj_id=item_to_patch, key=auth_key)
            success.append(item_to_patch)
        except Exception as e:
            errors[item_to_patch] = str(e)
    return success, errors


def add_to_patch(patch_dict, item_properties, meta_workflow_run_uuid):
    """Add item UUID and updated associated_meta_workflow_runs PATCH
    body to patch_dict.

    :param patch_dict: Dictionary with key, value pairs of item UUIDs
        and corresponding PATCH bodies
    :type patch_dict: dict
    :param item_properties: Item metadata
    :type item_properties: dict
    :param meta_workflow_run_uuid: MetaWorkflowRun UUID to add to
        item's associated_meta_workflow_runs
    :type meta_workflow_run_uuid: str
    """
    if item_properties:
        item_uuid = item_properties.get("uuid")
        item_associated_meta_workflow_runs = item_properties.get(
            "associated_meta_workflow_runs", []
        )
        item_associated_meta_workflow_runs.append(meta_workflow_run_uuid)
        if item_uuid:
            patch_dict[item_uuid] = {
                "associated_meta_workflow_runs": item_associated_meta_workflow_runs
            }



#!/usr/bin/env python3

################################################
#
#   Library to work with MetaWorkflowRun[json]
#
#   Michele Berselli
#   berselli.michele@gmail.com
#
################################################

################################################
#   Libraries
################################################
import sys, os
import copy

################################################
#   MetaWorkflowRun
################################################
class MetaWorkflowRun(object):
    """Class to represent a MetaWorkflowRun[json].
    """

    def __init__(self, input_json):
        """Constructor method.
        Initialize object and attributes.

        :param input_json: MetaWorkflowRun[json]
        :type input_json: dict
        """
        # Copy it so that the original does not get changed unexpectedly
        input_json_ = copy.deepcopy(input_json)

        # Basic attributes
        for key in input_json_:
            setattr(self, key, input_json_[key])
        #end for
        # Calculated attributes
        self.runs = {} #{run_obj.shard_name: run_obj, ...}

        # Calculate attributes
        self._validate()
        self._read_runs()
    #end def

    class WorkflowRun(object):
        """Class to represent a WorkflowRun[json].
        """

        def __init__(self, input_json):
            """Constructor method.
            Initialize object and attributes.

            :param input_json: WorkflowRun[json]
            :type input_json: dict
            """
            # Basic attributes
            for key in input_json:
                setattr(self, key, input_json[key])
            #end for
            if not getattr(self, 'output', None):
                self.output = []
            #end if
            if not getattr(self, 'dependencies', None):
                self.dependencies = []
            #end if
            # Validate
            self._validate()
            # Calculated attributes
            self.shard_name = self.name + ':' + self.shard
        #end def

        def _validate(self):
            """
            """
            try:
                getattr(self, 'name') #str
                getattr(self, 'status') #str, pending | running | completed | failed
                getattr(self, 'shard') #str
            except AttributeError as e:
                raise ValueError('JSON validation error, {0}\n'
                                    .format(e.args[0]))
            #end try
        #end def

    #end class

    def _validate(self):
        """
        """
        try:
            getattr(self, 'meta_workflow') #str
            getattr(self, 'input') #list
            getattr(self, 'workflow_runs') #list
            getattr(self, 'final_status') #str, pending | running | completed | failed
        except AttributeError as e:
            raise ValueError('JSON validation error, {0}\n'
                                .format(e.args[0]))
        #end try
    #end def

    def _read_runs(self):
        """
        """
        for run in self.workflow_runs:
            run_obj = self.WorkflowRun(run)
            if run_obj.shard_name not in self.runs:
                self.runs.setdefault(run_obj.shard_name, run_obj)
            else:
                raise ValueError('Validation error, step "{0}" duplicate in step workflows\n'
                                    .format(run_obj.shard_name))
            #end if
        #end for
    #end def

    def to_run(self):
        """Find all pending WorkflowRun[obj] that completed
        dependencies and are ready to run.

        :return: List of WorkflowRun[obj] that completed
            dependencies and are ready to run
        :rtype: list(object)
        """
        runs_ = []
        for _, run_obj in self.runs.items():
            if run_obj.status == 'pending':
                is_dependencies = True
                # Check dependencies are completed
                for shard_name_ in run_obj.dependencies:
                    if self.runs[shard_name_].status != 'completed':
                        is_dependencies = False
                        break
                    #end if
                #end for
                if is_dependencies: runs_.append(run_obj)
                #end if
            #end if
        #end for
        return runs_
    #end def

    def running(self):
        """Find all WorkflowRun[obj] that have status set to running.

        :return: List of WorkflowRun[obj] that have
            status set to running
        :rtype: list(object)
        """
        runs_ = []
        for _, run_obj in self.runs.items():
            if run_obj.status == 'running':
                runs_.append(run_obj)
            #end if
        #end for
        return runs_
    #end def

    def update_attribute(self, shard_name, attribute, value):
        """Update attribute value for WorkflowRun[obj] in runs.

        :param shard_name: WorkflowRun[obj] shard_name ('name:shard')
        :type shard_name: str
        :param attribute: attribute to update
        :type attribute: str
        :param value: new value for attribute
        """
        setattr(self.runs[shard_name], attribute, value)
    #end def

    def runs_to_json(self):
        """
        :return: List of dictionaries for workflow_runs
        :rtype: list(dict)
        """
        runs_ = []
        for run in self.workflow_runs: #used to get the right order
            run_ = {}
            shard_name = run['name'] + ':' + run['shard']
            for key, val in vars(self.runs[shard_name]).items():
                if val and key != 'shard_name':
                    run_.setdefault(key, val)
                #end if
            #end for
            runs_.append(run_)
        #end for
        return runs_
    #end def

    def to_json(self):
        """
        :return: MetaWorkflowRun[json]
        :rtype: dict
        """
        run_json = {}
        # Get attributes
        for key, val in vars(self).items():
            if key != 'runs':
                run_json.setdefault(key, val)
            #end if
        #end for
        # Get updated workflow_runs from current WorkflowRun objects
        run_json['workflow_runs'] = self.runs_to_json()
        return run_json
    #end def

    def _reset_run(self, shard_name):
        """Reset attributes value for WorkflowRun[obj] in runs.

        :param shard_name: WorkflowRun[obj] shard_name ('name:shard')
        :type shard_name: str
        """
        run_obj = self.runs[shard_name]
        # Reset run_obj
        run_obj.output = []
        run_obj.status = 'pending'
        if getattr(run_obj, 'jobid', None):
            delattr(run_obj, 'jobid')
        #end if
    #end def

    def reset_step(self, step_name):
        """Reset attributes value for WorkflowRun[obj] in runs.
        Reset all WorkflowRun[obj] corresponding to step specified by step_name.

        :param step_name: Name of the step to reset
        :type step_name: str
        """
        for shard_name_ in self.runs:
            if step_name == shard_name_.split(':')[0]:
                # Reset run_obj
                self._reset_run(shard_name_)
            #end if
        #end for
    #end def

    def reset_shard(self, shard_name):
        """Reset attributes value for WorkflowRun[obj] in runs.
        Reset only WorkflowRun[obj] specified by shard_name.

        :param shard_name: WorkflowRun[obj] shard_name ('name:shard')
        :type shard_name: str
        """
        for shard_name_ in self.runs:
            if shard_name == shard_name_:
                # Reset run_obj
                self._reset_run(shard_name_)
            #end if
        #end for
    #end def

    def update_status(self): # failed -> running -> completed
        """Check status for all WorkflowRun[obj].
        If at least one is failed, set final_status as failed.
        If no failed and at least one is running, set final_status as running.
        If all are completed set final_status as completed.

        :return: final_status
        :rtype: str
        """
        self.final_status = 'pending' # initial set to pending
        all_completed = True
        is_completed = False
        for _, run_obj in self.runs.items():
            if run_obj.status != 'completed':
                all_completed = False
                if run_obj.status == 'failed':
                    self.final_status = 'failed'
                    break
                elif run_obj.status == 'running':
                    self.final_status = 'running'
                #end if
            else: is_completed = True
            #end if
        #end for
        if all_completed:
            self.final_status = 'completed'
        elif self.final_status == 'pending' and is_completed:
            self.final_status = 'inactive'
        #end if
        return self.final_status
    #end def

#end class



#!/usr/bin/env python3

################################################
#
#   dcicutils wrapper
#
################################################

################################################
#   Libraries
################################################

from magma.checkstatus import AbstractCheckStatus
from magma_smaht.wfrutils import FFWfrUtils


################################################
#   CheckStatusSMA
################################################
class CheckStatusSMA(AbstractCheckStatus):
    """Customized CheckStatus class for the portal.
    """

    def __init__(self, wflrun_obj, env=None):
        """Initialize the object and set all attributes.

        :param wflrun_obj: MetaWorkflowRun[obj]
        :type wflrun_obj: object
        :param env: Name of the environment to use (e.g. fourfront-cgap)
        :type env: str
        """
        super().__init__(wflrun_obj)

        # Portal-related attributes
        self._env = env
        # Cache for FFWfrUtils object
        self._ff = None
    #end def

    @property
    def status_map(self):
        """Mapping from get_status output to magma status.
        """
        return {
            'started': 'running',
            'complete': 'completed',
            'error': 'failed'
        }

    def check_running(self):
        """
        """
        
        for patch_dict in super().check_running():
            if patch_dict:
                failed_jobs = self.wflrun_obj.update_failed_jobs()
                if len(failed_jobs) > 0:
                    patch_dict['failed_jobs'] = failed_jobs
                yield patch_dict

    # The following three functions are for portal (cgap / 4dn)
    def get_uuid(self, job_id):
        """
        """
        return self.ff.wfr_run_uuid(job_id)

    def get_status(self, job_id):
        """
        """
        return self.ff.wfr_run_status(job_id)

    def get_output(self, job_id):
        """
        """
        return self.ff.get_minimal_processed_output(job_id)

    @property
    def ff(self):
        """Internal property used for get_status, get_output for portal.
        """
        if not self._ff:
            self._ff = FFWfrUtils(self._env)
        return self._ff

#end class



#!/usr/bin/env python3

################################################
#
#   Library to work with MetaWorkflow[json]
#
#   Michele Berselli
#   berselli.michele@gmail.com
#
################################################

################################################
#   Libraries
################################################
import sys, os
import copy

################################################
#   MetaWorkflow
################################################
class MetaWorkflow(object):
    """Class to represent a MetaWorkflow[json].
    """

    def __init__(self, input_json):
        """Constructor method.
        Initialize object and attributes.

        :param input_json: MetaWorkflow[json]
        :type input_json: dict
        """

        # Copy it so that the original does not get changed unexpectedly
        input_json_ = copy.deepcopy(input_json)

        # Basic attributes
        for key in input_json_:
            setattr(self, key, input_json_[key])
        #end for
        # Calculated attributes
        self.steps = {} #{step_obj.name: step_obj, ...}
        self._end_workflows = None

        # Calculate attributes
        self._validate()
        self._read_steps()

    #end def

    class StepWorkflow(object):
        """Class to represent a StepWorkflow[json]
        that is step of a MetaWorkflow[json].
        """

        def __init__(self, input_json):
            """Constructor method.
            Initialize object and attributes.

            :param input_json: StepWorkflow[json]
            :type input_json: dict
            """
            # Basic attributes
            for key in input_json:
                setattr(self, key, input_json[key])
            #end for
            # Calculated attributes
            self.is_scatter = 0 #dimension to scatter, int
            self.gather_from = {} #{name: dimension, ...} of steps to gather from
                                  # dimension is input argument dimension increment
                                  # and shard dimension decrement, int
            self.gather_input = {} #equivalent to gather_from but this will only
                                   # affect dependencies input and not the
                                   # scatter structure of the step
            # For building graph structure
            self._nodes = set() #step_objects for steps that depend on current step
            # Dependencies
            #   names of steps that are dependency
            if getattr(self, 'dependencies', None):
                self.dependencies = set(self.dependencies)
            else:
                self.dependencies = set()
            #end def

            # Calculate attributes
            self._validate()
            self._attributes()
            # Validate that gather and gather_input have not been set together
            #   for the step
            if self.gather_from and self.gather_input:
                raise ValueError('JSON validation error, {0}\n'
                    .format('gather and gather_input can\'t be used together in the same step'))
            #end if
        #end def

        def _validate(self):
            """
            """
            try:
                getattr(self, 'name') #str, need to be unique
                getattr(self, 'workflow') #str, need to be unique
                getattr(self, 'config') #dict
                getattr(self, 'input') #list
            except AttributeError as e:
                raise ValueError('JSON validation error, {0}\n'
                                    .format(e.args[0]))
            #end try
        #end def

        def _attributes(self):
            """Read arguments and set calculated attributes.
            """
            for arg in self.input:
                scatter = arg.get('scatter') #scatter dimension
                if scatter and scatter > self.is_scatter: #get max scatter
                    self.is_scatter = scatter
                #end if
                source = arg.get('source') #source step name
                if source:
                    self.dependencies.add(source)
                    gather = arg.get('gather')
                    if gather:
                        self.gather_from.setdefault(source, gather)
                    else:
                        gather_input = arg.get('gather_input')
                        if gather_input:
                            self.gather_input.setdefault(source, gather_input)
                        #end if
                    #end if
                #end if
            #end for
        #end def

    #end class

    def _validate(self):
        """
        """
        try:
            getattr(self, 'uuid') #str, need to be unique
            getattr(self, 'input') #list
            getattr(self, 'workflows') #list
        except AttributeError as e:
            raise ValueError('JSON validation error, {0}\n'
                                .format(e.args[0]))
        #end try
    #end def

    def _read_steps(self):
        """
        """
        for wfl in self.workflows:
            step_obj = self.StepWorkflow(wfl)
            if step_obj.name not in self.steps:
                self.steps.setdefault(step_obj.name, step_obj)
            else:
                raise ValueError('Validation error, step "{0}" duplicate in step workflows\n'
                                    .format(step_obj.name))
            #end if
        #end for
    #end def

    def _build_run(self, end_steps):
        """Build graph structure for MetaWorkflow[json] given end_steps.
        Backtrack from end_steps to link StepWorkflow[obj] that are dependencies.
        Return a set containg StepWorkflow[obj] that are entry point.

        :param end_steps: List of names for finals StepWorkflow[obj] to
            use while building the graph structure
        :type end_steps: list(str)
        :return: Set of StepWorkflow[obj] that are starting points
        :rtype: set(obj)
        """
        steps_ = set() #steps that are entry point to wfl_run
        for end_step in end_steps:
            # Initialize queue with end_step
            queue = [self.steps[end_step]]
            # Reconstructing dependencies
            while queue:
                step_obj = queue.pop(0)
                if step_obj.dependencies:
                    for dependency in step_obj.dependencies:
                        try:
                            queue.append(self.steps[dependency])
                            self.steps[dependency]._nodes.add(step_obj)
                        except Exception:
                            raise ValueError('Validation error, missing dependency step "{0}" in step workflows\n'
                                                .format(dependency))
                        #end try
                    #end for
                else: steps_.add(step_obj)
                #end if
            #end while
        #end for
        return steps_
    #end def

    def _order_run(self, end_steps):
        """Sort and list all StepWorkflow[obj]
        necessary to run MetaWorkflow[json] given end_steps.

        The function will:
            - _build_run to build a graph structure for MetaWorkflow[json]
            - navigate the graph structure starting from StepWorkflow[obj]
                that are entry points

        :param end_steps: List of names for finals StepWorkflow[obj] to
            use while building the graph structure
        :type end_steps: list(str)
        :return: List with sorted StepWorkflow[obj]
        :rtype: list(object)
        """
        steps_ = []
        queue = list(self._build_run(end_steps))
        while queue:
            step_obj = queue.pop(0)
            # Adding next steps to queue
            for node in step_obj._nodes:
                if node not in steps_ and node not in queue:
                    queue.append(node)
                #end if
            #end for
            # Checking if dependencies are satisfied already and step can be added to steps_
            is_dependencies = True
            if step_obj.dependencies:
                for dependency in step_obj.dependencies:
                    if self.steps[dependency] not in steps_:
                        is_dependencies = False
                        queue.append(step_obj)
                        break
                    #end if
                #end for
            #end if
            if is_dependencies:
                steps_.append(step_obj)
            #end if
        #end while
        return steps_
    #end def

    def _input_dimensions(self, input_structure):
        """Given input_structure as list calculate dimensions.

        :param input_structure: Structure for the input with maximum scatter as list
        :type input_structure: list
        :return: Input dimensions
        :rtype: dict
        """
        input_dimensions = {}
        input_dimensions.setdefault(1, [len(input_structure)])
        if isinstance(input_structure[0], list):
            input_dimensions.setdefault(2, [])
            for i in input_structure:
                input_dimensions[2].append(len(i))
                if isinstance(i[0], list):
                    input_dimensions.setdefault(3, [])
                    d_ = []
                    for ii in i:
                        d_.append(len(ii))
                    #end for
                    input_dimensions[3].append(d_)
                #end if
            #end for
        #end if
        return input_dimensions
    #end def

    def _shards(self, input_dimensions, dimension):
        """Given input_dimensions calculate shards for specified dimension.

        :param input_dimensions: Input dimensions
        :type input_dimensions: dict
        :param dimension: Dimension to calculate shards for
        :type dimension: int [1|2|3]
        :return: List of shards
        :rtype: list(str)
        """
        shards = []
        input_dimension = input_dimensions[dimension]
        if dimension == 1: #1st dimension
            for i in range(input_dimension[0]):
                shards.append([str(i)])
            #end for
        elif dimension == 2: #2nd dimension
            for i, d in enumerate(input_dimension):
                for ii in range(d):
                    shards.append([str(i), str(ii)])
                #end for
            #end for
        else: #3rd dimension
            for i, d in enumerate(input_dimension):
                for ii, dd in enumerate(d):
                    for iii in range(dd):
                        shards.append([str(i), str(ii), str(iii)])
                    #end for
                #end for
            #end for
        #end if
        return shards
    #end def

    def _shards_dimension(self, shards):
        """
        """
        return len(shards[0])

    def write_run(self, input_structure, end_steps=[]):
        """Create MetaWorkflowRun[json] for MetaWorkflow[json]
        given end_steps and input_structure.

        The function will:
            - _order_run to sort and list all necessary StepWorkflow[json]
            - use scatter, gather_from and dependencies information
                to create and collect shards for individual StepWorkflow[json]
            - complete attributes and other metadata for MetaWorkflow[json]

        :param end_steps: List of names for finals StepWorkflow[obj] to
            use while creating MetaWorkflowRun[json], if no end_steps
            is specified calculate using end_workflows
        :type end_steps: list(str)
        :param input_structure: Structure for the input with
            maximum scatter as list (e.g. [[A, B], [C, D], [E]])
        :type input_structure: list [1|2|3 dimensions]
        :return: MetaWorkflowRun[json]
        :rtype: dict
        """
        # Get end_steps
        if not end_steps:
            end_steps = self.end_workflows
        #end if
        # Make input_structure a list if is string
        if isinstance(input_structure, str):
            input_structure = [input_structure]
        #end if
        scatter = {} #{step_obj.name: dimension, ...}
        fixed_shards = {} #{step_obj.name: shards, ...}
        dimensions = self._input_dimensions(input_structure)
        steps_ = self._order_run(end_steps)
        run_json = {
            'meta_workflow': self.uuid,
            'workflow_runs': [],
            'input': [],
            'final_status': 'pending'
        }
        for step_obj in steps_:
            run_step = {}
            run_step.setdefault('name', step_obj.name)
            run_step.setdefault('status', 'pending')
            # Check scatter
            #   If is_scatter or dependency in scatter
            #       but not in gather_from
            #       current step must be scattered
            scatter_dimension = 0 #dimension to scatter if any
            if step_obj.is_scatter:
                scatter_dimension = step_obj.is_scatter
                # Check if higher dimension in scatter
                #   get max scatter
                for dependency in step_obj.dependencies:
                    if dependency in scatter and scatter[dependency] > scatter_dimension:
                        scatter_dimension = scatter[dependency]
                    #end if
                #end for
                scatter.setdefault(step_obj.name, scatter_dimension)
            else:
                in_gather, gather_dimensions = True, []
                for dependency in step_obj.dependencies:
                    if dependency in scatter:
                        scatter_dimension = scatter[dependency]
                        if dependency not in step_obj.gather_from:
                            in_gather = False
                            break
                        else:
                            gather_dimension = scatter_dimension - step_obj.gather_from[dependency]
                            gather_dimensions.append(gather_dimension)
                        #end if
                    #end if
                #end for
                if in_gather and gather_dimensions:
                    scatter_dimension = max(gather_dimensions)
                #end if
                if scatter_dimension > 0:
                    scatter.setdefault(step_obj.name, scatter_dimension)
                #end if
            #end if
            # Created shards
            #   Check if there are fixed shards,
            #       else calculate based in input, scatter, gather dimensions
            if hasattr(step_obj, 'shards'):
                shards = step_obj.shards
                fixed_shards.setdefault(step_obj.name, shards)
            elif scatter_dimension:
                shards = self._shards(dimensions, scatter_dimension)
            else: shards = [['0']] #no scatter, only one shard
            #end if
            for s in shards:
                run_step_ = copy.deepcopy(run_step)
                run_step_.setdefault('shard', ':'.join(s))
                for dependency in sorted(step_obj.dependencies):
                    run_step_.setdefault('dependencies', [])
                    # Check gather
                    #   If dependency in gather_from or gather_input,
                    #       dependencies must be aggregated from scatter
                    gather_from_ = None
                    if dependency in step_obj.gather_from:
                        gather_from_ = step_obj.gather_from
                    elif dependency in step_obj.gather_input:
                        gather_from_ = step_obj.gather_input
                    #end if
                    if gather_from_:
                        # Check if the previous step is fixed_shards
                        #   if so get shards from there
                        if dependency in fixed_shards:
                            shards_gather = fixed_shards[dependency]
                        else:
                            # If the previous step is NOT in fixed_shards
                            #   get shards for original scatter dimension
                            shards_gather = self._shards(dimensions, scatter[dependency])
                        #end if
                        # Reducing dimension organically to gather
                        gather_dimension = self._shards_dimension(shards_gather) - gather_from_[dependency]
                        for s_g in shards_gather:
                            if scatter_dimension == 0 or \
                                scatter_dimension > gather_dimension: #gather all from that dependency
                                run_step_['dependencies'].append('{0}:{1}'.format(dependency, ':'.join(s_g)))
                            elif s_g[:scatter_dimension] == s: #gather only corresponding subset
                                run_step_['dependencies'].append('{0}:{1}'.format(dependency, ':'.join(s_g)))
                            #end if
                        #end for
                    else:
                        run_step_['dependencies'].append('{0}:{1}'.format(dependency, ':'.join(s)))
                    #end if
                #end for
                run_json['workflow_runs'].append(run_step_)
            #end for
        #end for
        return run_json
    #end def

    @property
    def end_workflows(self):
        if not self._end_workflows:
            all_wfls = [wfl.get('name') for wfl in self.workflows]
            sources = []
            for wfl in self.workflows:
                sources.extend([arg.get('source') for arg in wfl.get('input')])
                sources.extend(wfl.get('dependencies', []))
            #end for
            self._end_workflows = list(set(all_wfls).difference(set(sources)))
        #end if
        return sorted(self._end_workflows)
    #end def

#end class



################################################
#
#   Function to check and patch status
#       runs in MetaWorkflowRun[portal]
#
################################################
from typing import Any, Dict

from dcicutils import ff_utils

from magma_ff import checkstatus
from magma_ff.metawflrun import MetaWorkflowRun
from magma_ff.utils import check_status, make_embed_request


################################################
#   Functions
################################################
def status_metawfr(
    metawfr_uuid, ff_key, verbose=False, env="fourfront-cgap", valid_status=None
):
    """Perform status check on MetaWorkflowRun[portal].

    Retrieve the MetaWorkflowRun[portal], check all of its runs for
    updates, check for QC failures (if applicable),
    and then PATCH MetaWorkflowRun[portal] with updates (if found).

    Calculate cost and include in PATCH body only if the MetaWorkflowRun
    has finished to save on number of calls to tibanna API.

    :param metawfr_uuid: MetaWorkflowRun[portal] UUID
    :type metawfr_uuid: str
    :param ff_key: Portal authorization key
    :type ff_key: dict
    :param env: Environment name
    :type env: str
    :param verbose: Whether to print the POST response
    :type verbose: bool
    :param valid_status: Status considered valid for MetaWorkflowRun
        final_status property
    :type valid_status: list(str) or None
    """
    perform_action = True
    patch_body = None
    run_json = ff_utils.get_metadata(
        metawfr_uuid, add_on="frame=raw&datastore=database", key=ff_key
    )
    perform_action = check_status(run_json, valid_status)
    if perform_action:
        ignore_quality_metrics = run_json.get("ignore_output_quality_metrics")
        run_obj = MetaWorkflowRun(run_json)
        cs_obj = checkstatus.CheckStatusFF(run_obj, env)
        status_updates = list(cs_obj.check_running())  # Get all updates
        if status_updates:
            patch_body = status_updates[-1]  # Take most updated
        if patch_body:
            if not ignore_quality_metrics:
                updated_workflow_runs = get_recently_completed_workflow_runs(
                    run_json, patch_body
                )
                quality_metrics_failing = evaluate_quality_metrics(
                    updated_workflow_runs, ff_key
                )
                if quality_metrics_failing:
                    patch_body["final_status"] = "quality metric failed"
            if is_final_status_completed(patch_body):
                patch_body["cost"] = run_obj.update_cost()
            patch_response = ff_utils.patch_metadata(
                patch_body, metawfr_uuid, key=ff_key
            )
            if verbose:
                print(patch_response)


def get_recently_completed_workflow_runs(meta_workflow_run, updated_properties):
    """Compare between original and updated MetaWorkflowRun[json] to find
    UUIDs for all newly finished runs.

    :param meta_workflow_run: Original MetaWorkflowRun[json]
    :type meta_workflow_run: dict
    :param updated_properties: Updated MetaWorkflowRun[json]
    :type updated_properties: dict
    :return: UUIDs of newly completed runs
    :rtype: list(str)
    :raises ValueError: If number of workflow runs on MetaWorkflowRun
        differs from that on the updated properties
    """
    result = []
    original_workflow_runs = meta_workflow_run.get("workflow_runs", [])
    updated_workflow_runs = updated_properties.get("workflow_runs", [])
    if len(original_workflow_runs) != len(updated_workflow_runs):
        raise ValueError(
            "Workflow run length unexpectedly changed during status update."
            "\nOriginal: %s\nUpdated: %s"
            % (original_workflow_runs, updated_workflow_runs)
        )
    for idx in range(len(original_workflow_runs)):
        original_workflow_run = original_workflow_runs[idx]
        updated_workflow_run = updated_workflow_runs[idx]
        original_status = original_workflow_run.get("run_status")
        updated_status = updated_workflow_run.get("run_status")
        if updated_status != original_status and updated_status == "completed":
            workflow_run_item = updated_workflow_run.get("workflow_run")
            if workflow_run_item:
                result.append(workflow_run_item)
    return result


def evaluate_quality_metrics(workflow_runs_to_check, ff_key):
    """Retrieve runs and evaluate their output QualityMetrics
    for any failures.

    :param workflow_runs_to_check: WorkflowRun[portal] UUIDs to evaluate
    :type workflow_runs_to_check: list(str)
    :param ff_key: Portal authorization key
    :type ff_key: dict
    :return: Whether any of given WorkflowRuns have failing QCs
    :rtype: bool
    """
    result = False
    embed_fields = ["output_files.value_qc.overall_quality_status"]
    embed_response = make_embed_request(workflow_runs_to_check, embed_fields, ff_key)
    for workflow_run in embed_response:
        quality_metrics_failed = evaluate_workflow_run_quality_metrics(workflow_run)
        if quality_metrics_failed:
            result = True
            break
    return result


def evaluate_workflow_run_quality_metrics(workflow_run):
    """Evaluate WorkflowRun[json] output QualityMetrics for failure.

    NOTE: All status other than FAIL (such as WARN) are considered
    equivalent to PASS here.

    :param workflow_run: WorkflowRun[json] to evaluate
    :type workflow_run: dict
    :return: Whether any ouput QualityMetric has failed
    :rtype: bool
    """
    result = False
    output_files = workflow_run.get("output_files", [])
    for output_file in output_files:
        quality_metric = output_file.get("value_qc", {})
        if quality_metric:
            quality_metric_status = quality_metric.get("overall_quality_status")
            if quality_metric_status == "FAIL":
                result = True
                break
    return result


def is_final_status_completed(meta_workflow_run: Dict[str, Any]) -> bool:
    """Determine whether MetaWorkflowRun has finished."""
    status = meta_workflow_run.get("final_status")
    return status == "completed"



#!/usr/bin/env python3

################################################
#
#   InputGenerator ff
#
################################################

################################################
#   Libraries
################################################
import sys, os

# tibanna
from tibanna.utils import create_jobid

# magma
from magma.inputgenerator import InputGenerator as InputGeneratorFromMagma
from magma.inputgenerator import Argument

################################################
#   InputGenerator
################################################
class InputGenerator(InputGeneratorFromMagma):

    def __init__(self, wfl_obj, wflrun_obj):
        super().__init__(wfl_obj, wflrun_obj)

        # Output from ff use file as key
        #      instead of files for file argument value
        self.file_key = 'file'
    #end def

    def input_generator(self, env='env'):
        """Function customized for tibanna and portal interaction.
        Require env as an additional input.

        For each WorkflowRun[obj] ready to run:
            - Update status to 'running'
            - Add a job_id
            - Create specific input as dict
            - Create updated workflow_runs and final_status properties
            - Yield input, {workflow_runs, final_status}

        :param env: Environment to pass to tibanna (e.g. fourfront-cgap)
        :type env: str
        :return: Generator to input for WorkflowRun[obj] and updated
            workflow_runs and final_status information for patching
        :rtype: generator(dict, dict)
        """
        for run_obj, run_args in self._input():
            job_id = create_jobid()
            # Update run status and job_id
            self.wflrun_obj.update_attribute(run_obj.shard_name, 'status', 'running')
            self.wflrun_obj.update_attribute(run_obj.shard_name, 'job_id', job_id)
            ### This is where formatting happens,
            #       to change formatting just change this part
            step_obj = self.wfl_obj.steps[run_obj.name]
            input_json = {
                'app_name': run_obj.name,
                'workflow_uuid': step_obj.workflow,
                'config': step_obj.config,
                'parameters': {},
                'input_files': [],
                'jobid': job_id,
                "_tibanna": {
                  "run_type": run_obj.name,
                  "env": env
                }
            }

            ####################################################################
            # This should either come from run_obj or step_obj
            if getattr(step_obj, 'custom_pf_fields', None):
                input_json.setdefault('custom_pf_fields', step_obj.custom_pf_fields)
            ####################################################################

            ####################################################################
            # This should either come from wflrun_obj or wfl_obj
            if getattr(self.wflrun_obj, 'common_fields', None):
                input_json.setdefault('common_fields', self.wflrun_obj.common_fields)
            ####################################################################

            for arg_obj in run_args:
                if arg_obj.argument_type in ['parameter', 'QC ruleset']:
                    input_json['parameters'].setdefault(arg_obj.argument_name, arg_obj.value)
                else:
                    # Basic argument information
                    arg_ = {
                        'workflow_argument_name': arg_obj.argument_name,
                        'uuid': arg_obj.files
                    }
                    # Additional information
                    if getattr(arg_obj, 'mount', None):
                        arg_.setdefault('mount', arg_obj.mount)
                    #end if
                    if getattr(arg_obj, 'rename', None):
                        rname = arg_obj.rename
                        if isinstance(rname, str) and rname.startswith('formula:'):
                            frmla = rname.split('formula:')[-1]
                            rname = self._value_parameter(frmla, self.wflrun_obj.input)
                        #end if
                        arg_.setdefault('rename', rname)
                    #end if
                    if getattr(arg_obj, 'unzip', None):
                        arg_.setdefault('unzip', arg_obj.unzip)
                    #end if
                    input_json['input_files'].append(arg_)
                #end if
            #end for
            yield input_json, {'final_status':  self.wflrun_obj.update_status(),
                               'workflow_runs': self.wflrun_obj.runs_to_json()}
        #end for
    #end def

#end class



#!/usr/bin/env python3

################################################
#
#   Function to compute and patch the cost
#         of a MetaWorkflowRun[portal]
#
################################################

################################################
#   Libraries
################################################
import sys, os

# magma
from magma_ff.metawflrun import MetaWorkflowRun

# dcicutils
from dcicutils import ff_utils

################################################
#   Functions
################################################
def update_cost_metawfr(metawfr_uuid, ff_key, verbose=False):
    """Updates cost for MetaWorkflowRun[portal].

    :parm metawfr_uuid: MetaWorkflowRun[portal] UUIDs
    :type metawfr_uuid: str
    :param ff_key: Portal authorization key
    :type ff_key: dict
    :param verbose: Whether to print the POST response
    :type verbose: bool
    """
    # Get MetaWorkflowRun[json] from the portal
    run_json = ff_utils.get_metadata(metawfr_uuid, add_on='frame=raw&datastore=database', key=ff_key)
    # Create MetaWorkflowRun[obj]
    run_obj = MetaWorkflowRun(run_json)

    cost = run_obj.update_cost()

    if cost is not None and cost > 0:
        patch_dict = {'cost':  cost}
        res_post = ff_utils.patch_metadata(patch_dict, metawfr_uuid, key=ff_key)
        if verbose:
            print(res_post)
        #end if
    #end if
#end def



#!/usr/bin/env python3

################################################
#
#   Parser to handle compatibility between
#       magma and portal json formats
#
#   Michele Berselli
#   berselli.michele@gmail.com
#
################################################

################################################
#   Libraries
################################################
import sys, os
import json

################################################
#   ParserFF
################################################
class ParserFF(object):
    """
    """

    def __init__(self, input_json):
        """Initialize the object and set all attributes.

        :param input_json: MetaWorkflow[portal] or MetaWorkflowRun[portal]
        :type input_json: dict
        """
        self.in_json = input_json
    #end def

    def arguments_to_json(self):
        """Parse MetaWorkflow[portal] or MetaWorkflowRun[portal] stored as self.in_json.
        If input, convert and replace arguments in input from string to json.
        If workflows, for each step convert and replace arguments in input from string to json.
        """
        if self.in_json.get('input'):
            self._input_to_json(self.in_json['input'])
        #end if
        if self.in_json.get('workflows'):
            self._workflows_to_json(self.in_json['workflows'])
        #end if
        return self.in_json
    #end def

    def _workflows_to_json(self, workflows):
        """
        """
        for workflow in workflows:
            self._input_to_json(workflow['input'])
        #end for
    #end def

    def _input_to_json(self, input):
        """Loop through arguments in input and
        calls appropriate conversion function for argument_type.
        """
        for arg in input:
            if arg['argument_type'] == 'file':
                self._file_to_json(arg)
            else: # is parameter
                self._parameter_to_json(arg)
            #end if
        #end for
    #end def

    def _file_to_json(self, arg):
        """Convert file argument from string to json
        and replace the argument value (files).
        """
        if arg.get('files'):
            arg['files'] = self._files_to_json(arg['files'])
        #end if
    #end def

    def _files_to_json(self, files, sep=','):
        """Convert file argument from string to json.

        :param files: List of dictionaries representing files
            and information on their dimensional structure
            e.g. "files": [
                            {
                                "file": "A",
                                "dimension": "0"
                                # 0D missing, 1D X, 2D X:X, 3D X:X:X, ...
                            },
                            {
                               "file": "B",
                               "dimension": "1"
                            }
                        ]
        :type files: list(dict)
        :return: Converted argument value (files)
        :rtype: list(str)
        """
        list_ = []
        # Get max dimensions needed
        for file in files:
            dimension = file.get('dimension')
            if not dimension:
                return file.get('file')
            #end if
            dimension_ = list(map(int, dimension.split(sep)))
            # Expand input list based on dimensions
            self._init_list(list_, dimension_)
            # Add element
            tmp_list = list_
            for i in dimension_[:-1]:
                tmp_list = tmp_list[i]
            #end for
            tmp_list[dimension_[-1]] = file.get('file')
        #end for
        return list_
    #end def

    def _init_list(self, list_, dimension_):
        """ """
        tmp_list = list_
        for i in dimension_[:-1]:
            try: # index esist
                tmp_list[i]
            except IndexError: # create index
                for _ in range(i-len(tmp_list)+1):
                  tmp_list.append([])
            #end try
            tmp_list = tmp_list[i]
        #end for
        for _ in range(dimension_[-1]-len(tmp_list)+1):
            tmp_list.append(None)
        #end for
    #end def

    def _parameter_to_json(self, arg):
        """Convert parameter argument from string to json
        and replaces the argument value (value).

        possible values for value_type:
            json | string | integer | boolean | float
        """
        if not arg.get('value'):
            return
        #end if
        value = arg['value']
        value_type = arg['value_type']
        if value_type == 'json':
            value = json.loads(value)
        elif value_type == 'integer':
            value = int(value)
        elif value_type == 'float':
            value = float(value)
        elif value_type == 'boolean':
            if value.lower() == 'true':
                value = True
            else: value = False
            #end if
        #end if
        arg['value'] = value
        del arg['value_type']
    #end def

#end class



#!/usr/bin/env python3

################################################
#
#   Functions to create a MetaWorkflowRun
#
################################################

################################################
#   Libraries
################################################

import json, uuid
from dcicutils import ff_utils
import pprint

pp = pprint.PrettyPrinter(indent=2)

# magma
from magma_smaht.metawfl import MetaWorkflow
from magma_smaht.utils import (
    get_latest_mwf,
    get_file_set,
    get_library_from_file_set,
    get_sample_from_library,
    get_mwfr_file_input_arg,
    get_mwfr_parameter_input_arg,
)

# MetaWorkflow names are used to get the latest version.
# We assume that they don't change!
MWF_NAME_ILLUMINA = "Illumina_alignment_GRCh38"
MWF_NAME_ONT = "ONT_alignment_GRCh38"
MWF_NAME_PACBIO = "PacBio_alignment_GRCh38"
MWF_NAME_HIC = "Hi-C_alignment_GRCh38"
MWF_NAME_FASTQC = "Illumina_FASTQ_quality_metrics"

# Input argument names
INPUT_FILES_R1_FASTQ_GZ = "input_files_r1_fastq_gz"
INPUT_FILES_R2_FASTQ_GZ = "input_files_r2_fastq_gz"
INPUT_FILES_BAM = "input_files_bam"
INPUT_FILES_FASTQ_GZ = "input_files_fastq_gz"
SAMPLE_NAME = "sample_name"
LENGTH_REQUIRED = "length_required"
LIBRARY_ID = "library_id"

# Schema fields
COMMON_FIELDS = "common_fields"
UUID = "uuid"
SUBMISSION_CENTERS = "submission_centers"
SEQUENCING_CENTER = "sequencing_center"
CONSORTIA = "consortia"
FILE_SETS = "file_sets"
META_WORFLOW_RUN = "MetaWorkflowRun"
ACCESSION = "accession"
ALIASES = "aliases"


################################################
#   Functions
################################################


def mwfr_illumina_alignment(fileset_accession, length_required, smaht_key):
    """Creates a MetaWorflowRun item in the portal for Illumina alignment of submitted files within a fileset"""

    mwf = get_latest_mwf(MWF_NAME_ILLUMINA, smaht_key)
    print(f"Using MetaWorkflow {mwf[ACCESSION]} ({mwf[ALIASES][0]})")

    file_set = get_file_set(fileset_accession, smaht_key)
    mwfr_input = get_core_alignment_mwfr_input_from_readpairs(
        file_set, INPUT_FILES_R1_FASTQ_GZ, INPUT_FILES_R2_FASTQ_GZ, smaht_key
    )
    # Illumina specific input
    mwfr_input.append(get_mwfr_parameter_input_arg(LENGTH_REQUIRED, length_required))
    create_and_post_mwfr(
        mwf[UUID], file_set, INPUT_FILES_R1_FASTQ_GZ, mwfr_input, smaht_key
    )


def mwfr_pacbio_alignment(fileset_accession, smaht_key):
    """Creates a MetaWorflowRun item in the portal for PacBio alignment of submitted files within a fileset"""

    mwf = get_latest_mwf(MWF_NAME_PACBIO, smaht_key)
    print(f"Using MetaWorkflow {mwf[ACCESSION]} ({mwf[ALIASES][0]})")

    file_set = get_file_set(fileset_accession, smaht_key)
    mwfr_input = get_core_alignment_mwfr_input(file_set, INPUT_FILES_BAM, smaht_key)
    create_and_post_mwfr(mwf[UUID], file_set, INPUT_FILES_BAM, mwfr_input, smaht_key)


def mwfr_hic_alignment(fileset_accession, smaht_key):
    """Creates a MetaWorflowRun item in the portal for HI-C alignment of submitted files within a fileset"""

    mwf = get_latest_mwf(MWF_NAME_HIC, smaht_key)
    print(f"Using MetaWorkflow {mwf[ACCESSION]} ({mwf[ALIASES][0]})")

    file_set = get_file_set(fileset_accession, smaht_key)
    mwfr_input = get_core_alignment_mwfr_input_from_readpairs(
        file_set, INPUT_FILES_R1_FASTQ_GZ, INPUT_FILES_R2_FASTQ_GZ, smaht_key
    )
    create_and_post_mwfr(
        mwf[UUID], file_set, INPUT_FILES_R1_FASTQ_GZ, mwfr_input, smaht_key
    )


def mwfr_ont_alignment(fileset_accession, smaht_key):
    """Creates a MetaWorflowRun item in the portal for ONT alignment of submitted files within a fileset"""

    mwf = get_latest_mwf(MWF_NAME_ONT, smaht_key)
    print(f"Using MetaWorkflow {mwf[ACCESSION]} ({mwf[ALIASES][0]})")

    # Collect Input
    file_set = get_file_set(fileset_accession, smaht_key)
    library = get_library_from_file_set(file_set, smaht_key)
    sample = get_sample_from_library(library, smaht_key)

    # We are only retrieving the fastq files and get the bams from the derived_from property
    search_filter = f"?type=UnalignedReads&file_format.display_title=fastq_gz&file_sets.uuid={file_set[UUID]}"
    files_fastq = ff_utils.search_metadata(f"/search/{search_filter}", key=smaht_key)
    files_fastq.reverse()

    # Create files list for input args
    fastqs, bams = [], []
    for dim, file_fastq in enumerate(files_fastq):
        fastqs.append({"file": file_fastq[UUID], "dimension": f"{dim}"})
        bams.append(
            {"file": file_fastq["derived_from"][0][UUID], "dimension": f"{dim}"}
        )

    mwfr_input = [
        get_mwfr_file_input_arg(INPUT_FILES_FASTQ_GZ, fastqs),
        get_mwfr_file_input_arg(INPUT_FILES_BAM, bams),
        get_mwfr_parameter_input_arg(SAMPLE_NAME, sample[ACCESSION]),
        get_mwfr_parameter_input_arg(LIBRARY_ID, library[ACCESSION]),
    ]

    create_and_post_mwfr(
        mwf["uuid"], file_set, INPUT_FILES_FASTQ_GZ, mwfr_input, smaht_key
    )


def mwfr_fastqc(fileset_accession, smaht_key):

    file_set = get_file_set(fileset_accession, smaht_key)
    mwf = get_latest_mwf(MWF_NAME_FASTQC, smaht_key)
    print(f"Using MetaWorkflow {mwf[ACCESSION]} ({mwf[ALIASES][0]})")

    # Get unaligned R2 reads in the fileset that don't have already QC
    search_filter = f"?file_sets.uuid={file_set[UUID]}&type=UnalignedReads&file_format.display_title=fastq_gz&read_pair_number=R2&quality_metrics=No+value"
    
    files_to_run_r2 = ff_utils.search_metadata((f"search/{search_filter}"), key=smaht_key)
    files_to_run_r2.reverse()

    if len(files_to_run_r2) == 0:
        print(f"No files to run for search {search_filter}")
        return
    
    # Create files list for input args
    files_input = []
    for dim, file_r2 in enumerate(files_to_run_r2):
        files_input.append({"file": file_r2[UUID], "dimension": f"1,{dim}"})
        files_input.append({"file": file_r2["paired_with"][UUID], "dimension": f"0,{dim}"})
    files_input_list = sorted(files_input, key=lambda x: x['dimension'])

    mwfr_input = [get_mwfr_file_input_arg(INPUT_FILES_FASTQ_GZ, files_input_list)]
    create_and_post_mwfr(
        mwf[UUID], file_set, INPUT_FILES_FASTQ_GZ, mwfr_input, smaht_key
    )


def get_common_fields(file_set):

    sequencing_centers = file_set[SUBMISSION_CENTERS]
    if len(sequencing_centers) != 1:
        raise Exception(
            f"Exacted exactly one submission center for file set {file_set[ACCESSION]} expected but got {len(sequencing_centers)}"
        )

    common_fields = {SEQUENCING_CENTER: sequencing_centers[0]}

    return common_fields


def get_core_alignment_mwfr_input_from_readpairs(
    file_set, file_input_arg_1, file_input_arg_2, smaht_key
):

    library = get_library_from_file_set(file_set, smaht_key)
    sample = get_sample_from_library(library, smaht_key)

    # We are only retrieving the R2 reads and get the R1 read from the paired_with property
    search_filter = (
        f"?type=UnalignedReads&read_pair_number=R2&file_sets.uuid={file_set[UUID]}"
    )
    reads_r2 = ff_utils.search_metadata(f"/search/{search_filter}", key=smaht_key)
    reads_r2.reverse()

    # Create files list for input args
    files_r1, files_r2 = [], []
    for dim, file_r2 in enumerate(reads_r2):
        files_r2.append({"file": file_r2[UUID], "dimension": f"{dim}"})
        files_r1.append({"file": file_r2["paired_with"][UUID], "dimension": f"{dim}"})

    mwfr_input = [
        get_mwfr_file_input_arg(file_input_arg_1, files_r1),
        get_mwfr_file_input_arg(file_input_arg_2, files_r2),
        get_mwfr_parameter_input_arg(SAMPLE_NAME, sample[ACCESSION]),
        get_mwfr_parameter_input_arg(LIBRARY_ID, library[ACCESSION]),
    ]
    return mwfr_input


def get_core_alignment_mwfr_input(file_set, file_input_arg, smaht_key):

    library = get_library_from_file_set(file_set, smaht_key)
    sample = get_sample_from_library(library, smaht_key)

    search_filter = f"?type=UnalignedReads&file_sets.uuid={file_set[UUID]}"
    search_result = ff_utils.search_metadata(f"/search/{search_filter}", key=smaht_key)
    search_result.reverse()

    # Create files list for input args
    files = []
    for dim, file in enumerate(search_result):
        files.append({"file": file[UUID], "dimension": f"{dim}"})

    mwfr_input = [
        get_mwfr_file_input_arg(file_input_arg, files),
        get_mwfr_parameter_input_arg(SAMPLE_NAME, sample[ACCESSION]),
        get_mwfr_parameter_input_arg(LIBRARY_ID, library[ACCESSION]),
    ]
    return mwfr_input


def create_and_post_mwfr(mwf_uuid, file_set, input_arg, mwfr_input, smaht_key):

    mwfr = mwfr_from_input(mwf_uuid, mwfr_input, input_arg, smaht_key)
    mwfr[FILE_SETS] = [file_set[UUID]]
    mwfr[COMMON_FIELDS] = get_common_fields(file_set)
    #mwfr['final_status'] = 'stopped'
    #pprint.pprint(mwfr)

    post_response = ff_utils.post_metadata(mwfr, META_WORFLOW_RUN, smaht_key)
    mwfr_accession = post_response["@graph"][0]["accession"]
    print(
        f"Posted MetaWorkflowRun {mwfr_accession} for Fileset {file_set[ACCESSION]}."
    )


def filter_list_of_dicts(list_of_dics, property_target, target):
    return [w for w in list_of_dics if w[property_target] == target]


def mwfr_from_input(
    metawf_uuid,
    input,
    input_arg,
    ff_key,
    consortia=["smaht"],
    submission_centers=["smaht_dac"],
):
    """Create a MetaWorkflowRun[json] from the given MetaWorkflow[portal]
    and input arguments.

    :param metawf_uuid: MetaWorkflow[portal] UUID
    :type metawf_uuid: str
    :param input: Input arguments as list, where each argument is a dictionary
    :type list(dict)
    :param input_arg: argument_name of the input argument to use
        to calculate input structure
    :type str
    :param ff_key: Portal authorization key
    :type ff_key: dict

        e.g. input,
            input = [{
                    'argument_name': 'ARG_NAME',
                    'argument_type': 'file',
                    'files':[{'file': 'UUID', 'dimension': str(0)}]
                    }, ...]
    """

    metawf_meta = get_metadata(metawf_uuid, ff_key)

    for arg in input:
        if arg["argument_name"] == input_arg:
            input_structure = generate_input_structure(arg["files"])
        
    mwf = MetaWorkflow(metawf_meta)
    mwfr = mwf.write_run(input_structure)

    mwfr[UUID] = str(uuid.uuid4())
    mwfr[CONSORTIA] = consortia
    mwfr[SUBMISSION_CENTERS] = submission_centers
    mwfr["input"] = input

    return mwfr

def generate_input_structure(files):
    dimension_first_file = files[0]["dimension"] # We assume that this is representative of the input structure
    if dimension_first_file.count(",") == 0:
        return list(range(len(files)))
    elif dimension_first_file.count(",") == 1:
        dimensions = list(map(lambda x: x["dimension"].split(","), files))
        dimensions = list(map(lambda x: [int(x[0]), int(x[1])], dimensions))
        # Example for dimensions: [[1, 0],[0, 0],[1, 1],[0, 1],[1, 2]]
        dimensions_dict = {}
        for dim in dimensions:
            if dim[0] not in dimensions_dict:
                dimensions_dict[dim[0]] = [dim[1]]
            else:
                dimensions_dict[dim[0]].append(dim[1])
        # Example for dimensions_dict: {0: [0, 1], 1: [0, 1, 2]}
        input_structure = []
        for key in sorted(dimensions_dict.keys()):
            input_structure.append(dimensions_dict[key])
        # Example for input_structure: [[0, 1], [0, 1, 2]]
        return input_structure
    else:
        print("More than 2 input dimensions are currently no supported")
        exit()



def get_metadata(identifier, key):
    return ff_utils.get_metadata(
        identifier, add_on="frame=raw&datastore=database", key=key
    )



################################################
#
#   Functions to reset runs status
#       in MetaWorkflowRun[portal]
#
################################################
from dcicutils import ff_utils

from magma_smaht.metawflrun import MetaWorkflowRun
from magma_smaht.runupdate import RunUpdate
from magma_smaht.utils import check_status


################################################
#   Functions
################################################
def reset_steps(metawfr_uuid, steps_name, ff_key, verbose=False, valid_status=None):
    """Reset runs in MetaWorkflowRun[portal] corresponding
    to specified step names.
    PATCH MetaWorkflowRun[portal] with updates.

    :param metawfr_uuid: MetaWorkflowRun[portal] UUID
    :type metawfr_uuid: str
    :param steps_name: Step names for runs to reset
    :type steps_name: list(str)
    :param ff_key: Portal authorization key
    :type ff_key: dict
    :param verbose: Whether to print the POST response
    :type verbose: bool
    :param valid_status: Status considered valid for MetaWorkflowRun[portal]
        final_status property
    :type valid_status: list(str) or None
    """
    perform_action = True
    meta_workflow_run = ff_utils.get_metadata(
        metawfr_uuid, add_on="frame=raw&datastore=database", key=ff_key
    )
    perform_action = check_status(meta_workflow_run, valid_status)
    if perform_action:
        run_obj = MetaWorkflowRun(meta_workflow_run)
        runupd_obj = RunUpdate(run_obj)
        patch_dict = runupd_obj.reset_steps(steps_name)
        res_post = ff_utils.patch_metadata(patch_dict, metawfr_uuid, key=ff_key)
        if verbose:
            print(res_post)


def reset_shards(metawfr_uuid, shards_name, ff_key, verbose=False, valid_status=None):
    """Reset runs in MetaWorkflowRun[portal] corresponding
    to specified shard names.
    PATCH MetaWorkflowRun[portal] with updates.

    :param metawfr_uuid: MetaWorkflowRun[portal] UUID
    :type metawfr_uuid: str
    :param shards_name: Shard names for runs to reset
    :type shards_name: list(str)
    :param ff_key: Portal authorization key
    :type ff_key: dict
    :param verbose: Whether to print the POST response
    :type verbose: bool
    :param valid_status: Status considered valid for MetaWorkflowRun
        final_status property
    :type valid_status: list(str) or None
    """
    perform_action = True
    meta_workflow_run = ff_utils.get_metadata(
        metawfr_uuid, add_on="frame=raw&datastore=database", key=ff_key
    )
    perform_action = check_status(meta_workflow_run, valid_status)
    if perform_action:
        run_obj = MetaWorkflowRun(meta_workflow_run)
        runupd_obj = RunUpdate(run_obj)
        patch_dict = runupd_obj.reset_shards(shards_name)
        res_post = ff_utils.patch_metadata(patch_dict, metawfr_uuid, key=ff_key)
        if verbose:
            print(res_post)


def reset_status(
    metawfr_uuid, status, step_name, ff_key, verbose=False, valid_status=None
):
    """Reset runs in MetaWorkflowRun[portal] corresponding
    to specified step names if status match specified status.
    PATCH MetaWorkflowRun[portal] with updates.

    :param metawfr_uuid: MetaWorkflowRun[portal] UUID
    :type metawfr_uuid: str
    :param status: Status to reset
    :type status: str or list(str)
    :param steps_name: Step names for runs to reset
    :type steps_name: str or list(str)
    :param ff_key: Portal authorization key
    :type ff_key: dict
    :param verbose: Whether to print the POST response
    :type verbose: bool
    :param valid_status: Status considered valid for MetaWorkflowRun
        final_status property
    :type valid_status: list(str) or None
    """
    if isinstance(status, str):
        status = [status]
    if isinstance(step_name, str):
        step_name = [step_name]
    perform_action = True
    meta_workflow_run = ff_utils.get_metadata(
        metawfr_uuid, add_on="frame=raw&datastore=database", key=ff_key
    )
    perform_action = check_status(meta_workflow_run, valid_status)
    if perform_action:
        run_obj = MetaWorkflowRun(meta_workflow_run)
        to_reset = []
        for shard_name, obj in run_obj.runs.items():
            if obj.status in status and obj.name in step_name:
                to_reset.append(shard_name)
        runupd_obj = RunUpdate(run_obj)
        patch_dict = runupd_obj.reset_shards(to_reset)
        res_post = ff_utils.patch_metadata(patch_dict, metawfr_uuid, key=ff_key)
        if verbose:
            print(res_post)


def reset_all(metawfr_uuid, ff_key, verbose=False, valid_status=None):
    """Reset all runs in MetaWorkflowRun[portal].
    PATCH MetaWorkflowRun[portal] with updates.

    :param metawfr_uuid: MetaWorkflowRun[portal] UUID
    :type metawfr_uuid: str
    :param ff_key: Portal authorization key
    :type ff_key: dict
    :param verbose: Whether to print the POST response
    :type verbose: bool
    :param valid_status: Status considered valid for MetaWorkflowRun
        final_status property
    :type valid_status: list(str) or None
    """
    perform_action = True
    meta_workflow_run = ff_utils.get_metadata(
        metawfr_uuid, add_on="frame=raw&datastore=database", key=ff_key
    )
    perform_action = check_status(meta_workflow_run, valid_status)
    if perform_action:
        run_obj = MetaWorkflowRun(meta_workflow_run)
        to_reset = []
        for shard_name in run_obj.runs:
            to_reset.append(shard_name)
        runupd_obj = RunUpdate(run_obj)
        patch_dict = runupd_obj.reset_shards(to_reset)
        res_post = ff_utils.patch_metadata(patch_dict, metawfr_uuid, key=ff_key)
        if verbose:
            print(res_post)


def reset_failed(metawfr_uuid, ff_key, verbose=False, valid_status=None):
    """Reset all failed runs in MetaWorkflowRun[portal].
    PATCH MetaWorkflowRun[portal] with updates.

    :param metawfr_uuid: MetaWorkflowRun[portal] UUID
    :type metawfr_uuid: str
    :param ff_key: Portal authorization key
    :type ff_key: dict
    :param verbose: Whether to print the POST response
    :type verbose: bool
    :param valid_status: Status considered valid for MetaWorkflowRun[portal]
        final_status property
    :type valid_status: list(str) or None
    """
    perform_action = True
    meta_workflow_run = ff_utils.get_metadata(
        metawfr_uuid, add_on="frame=raw&datastore=database", key=ff_key
    )
    perform_action = check_status(meta_workflow_run, valid_status)
    if perform_action:
        run_obj = MetaWorkflowRun(meta_workflow_run)
        to_reset = []
        for shard_name, obj in run_obj.runs.items():
            if obj.status == "failed":
                to_reset.append(shard_name)
        runupd_obj = RunUpdate(run_obj)
        patch_dict = runupd_obj.reset_shards(to_reset)
        res_post = ff_utils.patch_metadata(patch_dict, metawfr_uuid, key=ff_key)
        if verbose:
            print(res_post)



#!/usr/bin/env python3

################################################
#
#   Object to combine and integrate
#       MetaWorkflow[obj] and MetaWorkflowRun[obj]
#       to generate input for WorfklowRun[obj]
#
#   Michele Berselli
#   berselli.michele@gmail.com
#
################################################

################################################
#   Libraries
################################################
import sys, os
import re

# tibanna
from tibanna.utils import create_jobid

################################################
#   Argument
################################################
class Argument(object):
    """Class to model an argument.
    """

    def __init__(self, input_json):
        """Constructor method.
        Initialize object and attributes.

        :param input_json: Argument in json format
            (e.g. from StepWorkflow[json])
        :type input_json: dict
        """
        # Basic attributes
        for key in input_json:
            setattr(self, key, input_json[key])
        #end for
        # Validate
        self._validate()
        # Calculated attributes
        if not getattr(self, 'source_argument_name', None):
            self.source_argument_name = self.argument_name
        #end if
    #end def

    def _validate(self):
        """
        """
        try:
            getattr(self, 'argument_name')
            getattr(self, 'argument_type')
        except AttributeError as e:
            raise ValueError('JSON validation error, {0}\n'
                                .format(e.args[0]))
        #end try
    #end def

#end class

################################################
#   InputGenerator
################################################
class InputGenerator(object):
    """Class to combine MetaWorkflow[obj] and MetaWorkflowRun[obj].
    Has methods to generate specific input for WorkflowRun[obj] and
    information for patching MetaWorkflowRun[obj].
    """

    def __init__(self, wfl_obj, wflrun_obj):
        """Constructor method.
        Initialize object and attributes.

        :parm wfl_obj: MetaWorkflow[obj] representing a MetaWorkflow[json]
        :type wfl_obj: object
        :parm wflrun_obj: MetaWorkflowRun[obj] representing a MetaWorkflowRun[json]
        :type wflrun_obj: object
        """
        # Basic attributes
        self.wfl_obj = wfl_obj
        self.wflrun_obj = wflrun_obj
        # Key to use to access file value information
        #   in workflow-runs output
        self.file_key = 'files'
    #end def

    def input_generator(self):
        """Template function, can be customized to any specific format.

        For each WorkflowRun[obj] ready to run in MetaWorkflowRun[obj]:
            - Update status to 'running'
            - Add a jobid
            - Create specific input as dict
            - Create updated workflow_runs and final_status properties
            - Yield input, {workflow_runs, final_status}

        :return: Generator to input for WorkflowRun[obj] and updated
            workflow_runs and final_status information for patching
        :rtype: generator(dict, dict)
        """
        for run_obj, run_args in self._input():
            jobid = create_jobid()
            # Update run status and jobid
            self.wflrun_obj.update_attribute(run_obj.shard_name, 'status', 'running')
            self.wflrun_obj.update_attribute(run_obj.shard_name, 'jobid', jobid)
            ### This is where formatting happens,
            #       to change formatting just change this part
            step_obj = self.wfl_obj.steps[run_obj.name]
            input_json = {
                'name': run_obj.name,
                'workflow': step_obj.workflow,
                'config': self._eval_formula(step_obj.config),
                'parameters': {},
                'input_files': [],
                'jobid': jobid
            }

            for arg_obj in run_args:
                if arg_obj.argument_type == 'parameter':
                    input_json['parameters'].setdefault(arg_obj.argument_name, arg_obj.value)
                else:
                    # Basic argument information
                    arg_ = {
                        'workflow_argument_name': arg_obj.argument_name,
                        'files': arg_obj.files
                    }
                    # Additional information
                    if getattr(arg_obj, 'mount', None):
                        arg_.setdefault('mount', arg_obj.mount)
                    #end if
                    if getattr(arg_obj, 'rename', None):
                        rname = arg_obj.rename
                        if isinstance(rname, str) and rname.startswith('formula:'):
                            frmla = rname.split('formula:')[-1]
                            rname = self._value_parameter(frmla, self.wflrun_obj.input)
                        #end if
                        arg_.setdefault('rename', rname)
                    #end if
                    if getattr(arg_obj, 'unzip', None):
                        arg_.setdefault('unzip', arg_obj.unzip)
                    #end if
                    input_json['input_files'].append(arg_)
                #end if
            #end for
            yield input_json, {'final_status':  self.wflrun_obj.update_status(),
                               'workflow_runs': self.wflrun_obj.runs_to_json()}
        #end for
    #end def

    def _eval_formula(self, dit):
        """Replace formulas in dit with corresponding calculated values.
        Detect formula as "formula:<formula>".

        :param dit: Dictionary to evaluate for formulas
        :type dit: dict
        """
        d_ = {}
        for k, v in dit.items():
            if isinstance(v, str) and v.startswith('formula:'):
                frmla = v.split('formula:')[-1]
                # match parameters
                re_match = re.findall('([a-zA-Z_]+)', frmla)
                # replace parameters
                for s in re_match:
                    try:
                        val = self._value_parameter(s, self.wflrun_obj.input)
                        frmla = frmla.replace(s, str(val))
                    except Exception:
                        pass
                    #end try
                #end for
                v = eval(frmla)
            #end if
            d_.setdefault(k, v)
        #end for
        return d_
    #end def

    def _input(self):
        """
        """
        out_ = []
        # Get workflow-runs that need to be run
        for run_obj in self.wflrun_obj.to_run():
            # Get workflow-run arguments
            run_args = self._run_arguments(run_obj)
            # Match and update workflow-run arguments
            #   file arguments -> files
            #   parameter arguments -> value
            self._match_arguments(run_args, run_obj)
            out_.append((run_obj, run_args))
        #end for
        return out_
    #end def

    def _run_arguments(self, run_obj):
        """
        :param run_obj: WorkflowRun[obj] representing a WorkflowRun[json]
        :type run_obj: object
        """
        run_args = []
        for arg in self.wfl_obj.steps[run_obj.name].input:
            arg_obj = Argument(arg)
            run_args.append(arg_obj)
        #end for
        return run_args
    #end def

    def _match_arguments(self, run_args, run_obj):
        """
        :param run_args: List of Argument objects
        :type run_args: list(object)
        :param run_obj: WorkflowRun[obj] representing a WorkflowRun[json]
        :type run_obj: object
        """
        for arg_obj in run_args:
            is_file = False
            # Check argument type
            if arg_obj.argument_type == 'file':
                is_file = True
                is_match = self._match_argument_file(arg_obj, run_obj)
            else: # is parameter
                is_match = self._match_argument_parameter(arg_obj)
            #end if
            if not is_match:
                raise ValueError('Value error, cannot find a match for argument "{0}"\n'
                                    .format(arg_obj.argument_name))
            #end if
            # Check scatter and input_dimension
            dimension_ = getattr(arg_obj, 'scatter', 0)
            dimension_ += getattr(arg_obj, 'input_dimension', 0)
            # input_dimension is additional dimension used to subset the input argument
            #   when creating the step specific input
            if dimension_:
                shard = map(int, run_obj.shard.split(':'))
                if is_file: in_ = arg_obj.files
                else: in_ = arg_obj.value
                #end if
                for idx in list(shard)[:dimension_]:
                    # [:dimension_] handle multiple scatter in same shard,
                    #   use scatter dimension to subset shard index list
                    in_ = in_[idx]
                #end for
                if is_file: arg_obj.files = in_
                else: arg_obj.value = in_
                #end if
            #end if
        #end for
    #end def

    def _match_argument_file(self, arg_obj, run_obj):
        """
        :param arg_obj: Argument object
        :type arg_obj: object
        :param run_obj: WorkflowRun[obj] representing a WorkflowRun[json]
        :type run_obj: object
        """
        if getattr(arg_obj, 'source', None):
        # Is workflow-run dependency, match to workflow-run output
            file_ = []
            for dependency in run_obj.dependencies:
                if arg_obj.source == dependency.split(':')[0]:
                    for arg in self.wflrun_obj.runs[dependency].output:
                        if arg_obj.source_argument_name == arg['argument_name']:
                            file_.append(arg[self.file_key])
                            break
                        #end if
                    #end for
                #end if
            #end for
            gather = getattr(arg_obj, 'gather', 0)
            # If not gather look for gather_input
            if not gather:
                gather += getattr(arg_obj, 'gather_input', 0)
            #end if
            gather += getattr(arg_obj, 'extra_dimension', 0)
            # extra_dimension is additional increment to dimension used
            #   when creating the step specific input
            if gather == 1:  # gather 1 dimension
                arg_obj.files = file_
            elif gather > 1:  # gather 2+ dimensions
                arg_obj.files = file_
                for i in range(1, gather):
                    arg_obj.files = [arg_obj.files]
            else:  # no gather
                arg_obj.files = file_[0]
            #end if
            return True
        else:
        # No dependency, match to general argument
            return self._match_general_argument(arg_obj)
        #end if
    #end def

    def _match_argument_parameter(self, arg_obj):
        """
        :param arg_obj: Argument object
        :type arg_obj: object
        """
        if getattr(arg_obj, 'value', None) != None:
            # Is value
            return True
        else:
        # No value, match to general argument
            return self._match_general_argument(arg_obj)
        #end if
    #end def

    def _match_general_argument(self, arg_obj):
        """
        :param arg_obj: Argument object
        :type arg_obj: object
        """
        # Try and match with meta-worfklow-run input
        if self._value(arg_obj, self.wflrun_obj.input):
            return True
        #end if
        # No match, try match to default argument in meta-worfklow
        if self._value(arg_obj, self.wfl_obj.input):
            return True
        #end if
        return False
    #end def

    def _value(self, arg_obj, arg_list):
        """
        :param arg_obj: Argument object
        :type arg_obj: object
        :param arg_list: List of arguments as dictionaries
        :type arg_list: list(dict)
        """
        for arg in arg_list:
            if arg_obj.source_argument_name == arg['argument_name'] and \
               arg_obj.argument_type == arg['argument_type']:
                if arg_obj.argument_type == 'file':
                    arg_obj.files = arg['files']
                else:
                    arg_obj.value = arg['value']
                #end if
                return True
            #end if
        #end for
        return False
    #end def

    def _value_parameter(self, arg_name, arg_list):
        """
        :param arg_name: Name of the argument
        :type arg_name: str
        :param arg_list: List of arguments as dictionaries to match
        :type arg_list: list(dict)
        """
        for arg in arg_list:
            if arg_name == arg['argument_name'] and \
               arg['argument_type'] == 'parameter':
                return arg['value']
            #end if
        #end for
        raise ValueError('Value error, cannot find a match for parameter "{0}"\n'
                            .format(arg_name))
    #end def

#end class



import argparse

from magma_ff.create_metawfr import create_meta_workflow_run
from magma_ff.utils import get_auth_key


def main() -> None:
    parser = argparse.ArgumentParser(description="Create MetaWorkflowRun")
    parser.add_argument(
        "input_item_identifier",
        help="Identifier for item (e.g. Sample) for input to MetaWorkflowRun",
    )
    parser.add_argument(
        "meta_workflow_identifier",
        help="MetaWorkflow identifier for input to MetaWorkflowRun",
    )
    parser.add_argument(
        "-e", "--auth-env", default="default", help="Env name in ~/.cgap-keys.json"
    )
    parser.add_argument(
        "--no-post", action="store_true", help="Do not POST the MetaWorkflowRun created"
    )
    parser.add_argument(
        "--no-patch",
        action="store_true",
        help="Do not PATCH the input item with the MetaWorkflowRun created",
    )
    args = parser.parse_args()
    auth_key = get_auth_key(args.auth_env)
    post = not args.no_post
    patch = not args.no_patch
    create_meta_workflow_run(
        args.input_item_identifier,
        args.meta_workflow_identifier,
        auth_key,
        post=post,
        patch=patch,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

################################################
#
#   Library to work with MetaWorkflowRun[obj]
#
#   Implement the template for a generic object
#       and generic methods for status checking
#
################################################

################################################
#   Libraries
################################################
import sys, os

################################################
#   AbstractCheckStatus
################################################
class AbstractCheckStatus(object):
    """Template for CheckStatus class.
    """

    def __init__(self, wflrun_obj):
        """Constructor method.
        Initialize object and attributes.

        :param wflrun_obj: MetaWorkflowRun[obj] representing a MetaWorkflowRun[json]
        :type wflrun_obj: object
        """
        # Basic attributes
        self.wflrun_obj = wflrun_obj

    @property
    def status_map(self):
        """Mapping from get_status output to magma status.
        Set to property so that inherited classes can overwrite it.
        """
        return {
            'pending': 'pending',
            'running': 'running',
            'completed': 'completed',
            'failed' : 'failed'
        }

    def check_running(self): # We can maybe have a flag that switch between tibanna or dcic utils functions
        """
        """
        for run_obj in self.wflrun_obj.running():

            jobid = run_obj.jobid if hasattr(run_obj, 'jobid') else run_obj.job_id
            # Check current status from jobid
            status_ = self.get_status(jobid)
            status = self.status_map[status_]

            # Update run status no matter what
            self.wflrun_obj.update_attribute(run_obj.shard_name, 'status', status)

            # Get run uuid
            run_uuid = self.get_uuid(jobid)

            # Update run uuid regardless of the status
            if run_uuid:  # some failed runs don't have run uuid
                self.wflrun_obj.update_attribute(run_obj.shard_name, 'workflow_run', run_uuid)

            if status == 'completed':

                # Get formatted output
                output = self.get_output(jobid)

                # Update output
                if output:
                    self.wflrun_obj.update_attribute(run_obj.shard_name, 'output', output)

            elif status == 'running':
                yield None  # yield None so that it doesn't terminate iteration
                continue
            else:  # failed
                # handle error status - anything to do before yielding the updated json
                self.handle_error(run_obj)
            #end if

            # Return the json to patch workflow_runs for both completed and failed
            #   and keep going so that it can continue updating status for other runs
            yield {'final_status':  self.wflrun_obj.update_status(),
                   'workflow_runs': self.wflrun_obj.runs_to_json()}
        #end for
    #end def

    # Inherited classes could define stuff to do with an error case
    def handle_error(self, run_obj):
        pass

    # Replace them with real functions for getting status or (formatted) output
    def get_status(self, jobid):
        """
        """
        return 'running'

    def get_output(self, jobid):
        """
        """
        return [{'argument_name': 'arg', 'files': 'uuid'}]

    def get_uuid(self, jobid):
        """
        """
        return 'uuid'

#end class



#!/usr/bin/env python3

################################################
#
#   MetaWorkflowRun ff
#
################################################

################################################
#   Libraries
################################################
import sys, os
import copy

# magma
from magma.metawflrun import MetaWorkflowRun as MetaWorkflowRunFromMagma
from magma_smaht.parser import ParserFF
from tibanna_ffcommon.core import API

################################################
#   MetaWorkflowRun
################################################
class MetaWorkflowRun(MetaWorkflowRunFromMagma):

    def __init__(self, input_json):
        """
        """
        input_json_ = copy.deepcopy(input_json)
        ParserFF(input_json_).arguments_to_json()

        super().__init__(input_json_)
    #end def

    def _reset_run(self, shard_name):
        """Reset attributes value for WorkflowRun[obj]
        specified by shard_name.

        :param shard_name: Name of the WorkflowRun[obj] to reset,
            include shard ('name:shard')
        :type shard_name: str
        """
        run_obj = self.runs[shard_name]
        # Reset run_obj
        run_obj.output = []
        run_obj.status = 'pending'
        if getattr(run_obj, 'job_id', None):
            delattr(run_obj, 'job_id')
        #end if
        if getattr(run_obj, 'workflow_run', None):
            delattr(run_obj, 'workflow_run')
        #end if
    #end def

    def update_failed_jobs(self):
        """Check status for all WorkflowRun[obj].
        Add failed ones to MetaWorflowRun[obj] attribute failed_jobs.

        :return: Failed jobs (failed_jobs)
        :rtype: list(str)
        """
        if not hasattr(self, 'failed_jobs'):
            self.failed_jobs = []
        #end if
        for _, run_obj in self.runs.items():
            if run_obj.status == 'failed':
                self.failed_jobs.append(run_obj.job_id)
            #end if
        #end for
        self.failed_jobs = list(set(self.failed_jobs)) # Remove duplicates
        return self.failed_jobs
    #end def

    def update_cost(self):
        """Compute an estimated cost for MetaWorkflowRun[obj].
        Include failed and completed runs.

        :return: MetaWorkflowRun[obj] cost
        :rtype: float
        """
        self.cost = 0.0
        if hasattr(self, 'failed_jobs'):
            for job_id in self.failed_jobs:
                try:
                    run_cost, _ = API().cost_estimate(job_id=job_id, force=True)
                    self.cost = self.cost + run_cost
                except Exception as e:
                    return 0.0
        #end for
        for _, run_obj in self.runs.items():
            # only look for completed jobs as failed ones are in self.failed_jobs
            if run_obj.status == 'completed':
                try:
                    run_cost, _ = API().cost_estimate(job_id=run_obj.job_id, force=True)
                    #print(f"Costs {run_obj.job_id}: {run_cost}")
                    self.cost = self.cost + run_cost
                except Exception:
                    # This can happen, if e.g. prices from AWS can't be retrieved
                    return 0.0
            #end if
        #end for
        return self.cost
    #end def

#end class
